import numpy as np 
import pandas as pd

class QLearningTable:
    def __init__(self,actions,learning_rate=0.05,reward_decay=0.9,e_greedy=0.1):
       
        self.actions=actions
        self.lr=learning_rate
        self.gamma=reward_decay
        self.epsilon=e_greedy

        self.q_table=pd.DataFrame(columns=self.actions,dtype=np.float64)                
    
    def choose_action(self,observation):
        self.check_state_exist(observation)
        
        #action selection
        if np.random.uniform()>self.epsilon:
            state_action =self.q_table.loc[observation,:]
            action =np.random.choice(state_action[state_action==np.max(state_action)].index)
        else:
            action = np.random.choice(self.actions)
        
        return action
    
    def learn(self,s,a,r,s_):
        self.check_state_exist(s)
        self.check_state_exist(s_)
        q_predict=self.q_table.loc[s,a]
        if s_!='Game_over' or s_!='Game_pass':
            q_target =r+self.gamma*self.q_table.loc[s_,:].max()
        else:
            q_target=r
        self.q_table.loc[s,a]+=self.lr*(q_target-q_predict)
    
    def check_state_exist(self,state):
        if state not in list(self.q_table.index):
            self.q_table=self.q_table.append(pd.Series([0]*len(self.actions),index=self.q_table.columns,name=state,))


from src.template.Player import Player


class TestPlayer(object):
    construction = {"_id": 0, "_no": 0, "x": 0, "y": 0, "width": 60, "height": 60}
    player = Player(construction)

    def test_get_xy_pos(self):
        assert type(self.player.get_xy_pos()) == tuple
        assert type(self.player.get_xy_pos()[0]) == int
        assert type(self.player.get_xy_pos()[1]) == int



"""
The cross language handler for C++
"""
import os
import os.path
import string
import random

from subprocess import PIPE, Popen

from mlgame.crosslang.exceptions import CompilationError

def compile_script(script_full_path):
    """
    Compile the script to an executable

    The exception will be raised when failed to compile the script.

    @param script_full_path The full path of the target script
    @return The execution command of the executable
    """
    lib_dir = os.path.join(os.path.dirname(__file__), "include")

    dir_path = os.path.dirname(script_full_path)
    main_script_file_path = _preprocess_script(script_full_path, dir_path)
    execute_file_path = os.path.join(dir_path, "ml_play.out")
    if os.path.exists(execute_file_path):
        os.remove(execute_file_path)

    compile_cmd = [
        "g++", main_script_file_path, "-I" + lib_dir, "--std=c++11",
        "-o", execute_file_path
    ]

    with Popen(compile_cmd, bufsize = 1,
        stdout = PIPE, stderr = PIPE, universal_newlines = True) as p:
        outs, errs = p.communicate()

    # Remove the generated script after compilation
    os.remove(main_script_file_path)

    if p.returncode != 0:
        raise CompilationError(os.path.basename(script_full_path), errs)

    return [execute_file_path]

def _preprocess_script(user_script_path, outfile_dir):
    """
    Append the content of user script to "cpp_include/main.cpp" and save to a new file

    @param user_script_path The path of the user script as one of the source
    @param outfile_dir The path of the directory to put the new file
    @return The path of the new file
    """
    basefile_path = os.path.join(os.path.dirname(__file__), "include", "base_main.cpp")
    char_choice = string.ascii_lowercase + string.digits
    outfile_name = ("main_" +
        "".join([random.choice(char_choice) for _ in range(8)]) +
        ".cpp")
    outfile_path = os.path.join(outfile_dir, outfile_name)

    with open(outfile_path, "w") as out_file, \
         open(basefile_path, "r") as in_file:
        # Include the user file
        out_file.write("#include \"{}\"\n".format(os.path.basename(user_script_path)))
        for line in in_file:
            out_file.write(line)

    return outfile_path



import random


class MLPlay:
    def __init__(self,ai_name:str,*args,**kwargs):
        self.other_cars_position = []
        self.coins_pos = []
        self.ai_name = ai_name
        self.speed = random.randint(9,11)*10
        print("Initial ml script")
        print(ai_name)
        print(kwargs)

    def update(self, scene_info: dict,*args,**kwargs):
        """
        Generate the command according to the received scene information
        """
        # print(scene_info)
        if scene_info["status"] != "GAME_ALIVE":
            return "RESET"

        if random.randint(1,100)<self.speed:
            return ["SPEED"]
        else:
            return []

    def reset(self):
        """
        Reset the status
        """
        self.speed = random.randint(7,11)*10

        # print("reset ml script")
        pass



import abc
from enum import Enum

from mlgame.argument.model import GroupAI
from mlgame.core.model import GameProgressSchema
from mlgame.utils.enum import get_ai_name
from mlgame.view.view_model import Scene


class GameResultState(str, Enum):
    """
    表示遊戲結束的狀態
    finish 表示遊戲有成功執行到最後，表示玩家通關，或是多個玩家至少一人通關
    fail 表示玩家闖關失敗，或是沒有任何一個玩家通關
    """
    FINISH = "finish"
    FAIL = "fail"
    NOT_YET = "not_yet"
    PASSED = "passed"
    UN_PASSED = "un_passed"
    RUNNING = "running"


class GameStatus():
    # TODO refactor
    """
        表示遊戲進行中的狀態
        GAME_ALIVE 表示遊戲進行中
        GAME_OVER 表示玩家闖關失敗，多人遊戲中，收到此狀態，表示輸掉此遊戲
        GAME_PASS 表示玩家闖關成功，多人遊戲中，收到此狀態，表示贏得此遊戲
    """
    GAME_ALIVE = "GAME_ALIVE"
    GAME_OVER = "GAME_OVER"
    GAME_PASS = "GAME_PASS"
    GAME_1P_WIN = "GAME_1P_WIN"
    GAME_2P_WIN = "GAME_2P_WIN"
    GAME_DRAW = "GAME_DRAW"


class PaiaGame(abc.ABC):
    def __init__(self, user_num: int=1,group_ai_list:list[GroupAI]=[], *args, **kwargs):
        self.scene = Scene(width=800, height=600, color="#4FC3F7", bias_x=0, bias_y=0)
        self.frame_count = 0
        self.game_result_state = GameResultState.FAIL
        self.status = GameStatus.GAME_ALIVE
        self.user_num = user_num
        self.game_mode = "NORMAL"
        self.ai_enabled = True
        self.game_state = None
        self.group_ai_list = group_ai_list
    @abc.abstractmethod
    def update(self, commands: dict):
        self.game_state.update(commands)
        self.frame_count += 1

    @abc.abstractmethod
    def get_data_from_game_to_player(self) -> dict:
        """
        send something to game AI
        we could send different data to different ai
        """
        data_to_player = {}
        data_to_1p = {
            "frame": self.frame_count,
            "status": self.status,
            "key": "value"

        }
        for i in range(self.user_num):
            data_to_player[get_ai_name(i)] = data_to_1p
        return data_to_player

    @abc.abstractmethod
    def reset(self):
        pass

    @abc.abstractmethod
    def get_scene_init_data(self) -> dict:
        # TODO add schema
        """
        Get the initial scene and object information for drawing on the web
        """
        # TODO add music or sound
        scene_init_data = {"scene": self.scene.__dict__,
                           "assets": [

                           ],
                           "background":[]
                           # "audios": {}
                           }
        return scene_init_data

    @abc.abstractmethod
    def get_scene_progress_data(self) -> GameProgressSchema:
        """
        Get the position of game objects for drawing on the web
        """

        scene_progress = {
            # background view data will be draw first
            "background": [],
            # game object view data will be draw on screen by order , and it could be shifted by WASD
            "object_list": [],
            "toggle": [],
            "toggle_with_bias": [],
            "foreground": [],
            # other information to display on web
            "user_info": [],
            # other information to display on web
            "game_sys_info": {}
        #     TODO add music and sound
        }
        return GameProgressSchema.model_validate(scene_progress)

    @abc.abstractmethod
    def get_game_result(self) -> dict:
        """
        send game result
        """
        return {"frame_used": self.frame_count,
                "result": {

                },

                }



class GameState(abc.ABC):
    def __init__(self,game:PaiaGame):
        self.game = game
        
    @abc.abstractmethod
    def update(self, *args, **kwargs):
        """
        Update the game state.
        """
        pass
    
    @abc.abstractmethod
    def get_scene_progress_data(self)->GameProgressSchema:
        """
        Get the scene progress data.
        """
        pass
    
    @abc.abstractmethod
    def reset(self):
        """
        Reset the game state.
        """
        pass


def get_paia_game_obj(game_cls, parsed_game_params: dict, user_num, group_ai_list:list[GroupAI]) -> PaiaGame:
    game = game_cls(user_num=user_num, **parsed_game_params, group_ai_list=group_ai_list)
    assert isinstance(game, PaiaGame), "Game " + str(game) + " should implement a abstract class : PaiaGame"
    return game



import math

import pygame

from mlgame.game.paia_game import PaiaGame
from mlgame.utils.enum import get_ai_name
from mlgame.view.audio_model import create_music_init_data, MusicProgressSchema
from mlgame.view.decorator import check_game_progress, check_game_result
from mlgame.view.view_model import create_text_view_data, create_asset_init_data, create_image_view_data, \
    create_line_view_data, Scene, create_polygon_view_data, create_rect_view_data, create_scene_progress_data
from .env import *
from .mazeMode import MazeMode
from .points import Check_point

'''need some fuction same as arkanoid which without dash in the name of fuction'''


class MazeCar(PaiaGame):
    def __init__(self, user_num, map_num, time_to_play, map_file=None, *args, **kwargs):
        super().__init__(user_num=user_num)
        # self.game_type = game_type
        self.user_num = user_num
        self.is_single = False
        if self.user_num == 1:
            self.is_single = True
        # self.maze_id = map_num - 1
        self.game_end_time = time_to_play
        # self.sensor_num = sensor_num
        self.sensor_num = 5

        if map_file is None:
            map_file = path.join(MAP_FOLDER, f"map_{map_num}.json")
        self.map_file = map_file
        self.game_mode = MazeMode(self.user_num, self.map_file, self.game_end_time, self.sensor_num)
        # self.game_mode.sound_controller.play_music()
        self.is_running = self.isRunning()
        self.map_width = self.game_mode.map.width
        self.map_height = self.game_mode.map.height
        self.scene = Scene(WIDTH, HEIGHT, BG_COLOR)
        self.origin_car_pos = [0, 0]

    # self.origin_car_pos = self.game_mode.car_info[0]["center"]

    def update(self, cmd_dict):
        # self.game_mode.ticks()
        self.frame_count += 1
        self.game_mode.handle_event()
        self.game_mode.update_sprite(cmd_dict)
        if not self.isRunning():
            self.is_running = False
            return "RESET"
        for car in self.game_mode.cars:
            if self.origin_car_pos != [0, 0]:
                break
            self.origin_car_pos = car.get_info()["center"]

    def get_data_from_game_to_player(self):
        scene_info = self.get_scene_info
        player_info = {}
        for car in self.game_mode.car_info:
            # type of car is dictionary
            player_info[get_ai_name(int(car["id"]))] = {"frame": scene_info["frame"],
                                                        "status": car["status"],
                                                        "x": car["coordinate"][0],
                                                        "y": car["coordinate"][1],
                                                        "angle": (car["angle"] * 180 / math.pi) % 360,
                                                        "R_sensor": car["r_sensor_value"]["distance"],
                                                        "L_sensor": car["l_sensor_value"]["distance"],
                                                        "F_sensor": car["f_sensor_value"]["distance"],
                                                        "L_T_sensor": car["l_t_sensor_value"]["distance"],
                                                        "R_T_sensor": car["r_t_sensor_value"]["distance"],
                                                        "end_x": self.game_mode.end_point.get_info()["coordinate"][0],
                                                        "end_y": self.game_mode.end_point.get_info()["coordinate"][1],
                                                        }
        return player_info

    def reset(self):
        self.frame_count = 0
        self.game_mode = MazeMode(self.user_num, self.map_file, self.game_end_time, self.sensor_num)
        # self.game_mode.sound_controller.play_music()

    def isRunning(self):
        return self.game_mode.isRunning()

    @property
    def get_scene_info(self):
        """
        Get the scene information
        """
        scene_info = {
            "frame": self.game_mode.frame,
        }

        for car in self.game_mode.car_info:
            # type of car is dictionary
            scene_info[str(car["id"]) + "P_position"] = car["topleft"]
        return scene_info

    def get_scene_init_data(self) -> dict:
        """
        Get the scene and object information for drawing on the web
        """
        game_info = {
            "scene": self.scene.__dict__,
            "background": [],
            "assets": [],
            "map_width": self.game_mode.map.tileWidth * 20,
            "map_height": self.game_mode.map.tileHeight * 20
        }
        game_info["assets"].append(create_asset_init_data("info", 300, 700, INFO_PATH, INFO_URL))
        game_info["assets"].append(create_asset_init_data("bg_img", 1000, 700, BG_PATH, BG_URL))

        game_info["assets"].append(create_asset_init_data("endpoint", 60, 60, ENDPOINT_PATH, ENDPOINT_URL))
        game_info["assets"].append(create_asset_init_data("checkpoint", 60, 60, CHECKPOINT_PATH, CHECKPOINT_URL))
        game_info["assets"].append(create_asset_init_data("checkpoint_2", 60, 60, CHECKPOINT2_PATH, CHECKPOINT_URL))

        for car in self.game_mode.car_info:
            file_path = path.join(ASSET_IMAGE_DIR, CARS_NAME[car["id"]])
            url = CARS_URL[car["id"]]
            car_init_info = create_asset_init_data("car_0" + str(car["id"] + 1), 50, 40, file_path, url)
            game_info["assets"].append(car_init_info)

        for wall in self.game_mode.walls:
            vertices = [(wall.body.transform * v) for v in wall.box.shape.vertices]
            vertices = [self.game_mode.trnsfer_box2d_to_pygame(v) for v in vertices]
            # game_info["background"].append(create_aapolygon_view_data("wall", vertices, WHITE))
            game_info["background"].append(create_polygon_view_data("wall", vertices, WALL_COLOR))
            # game_info["background"].append(
            #     create_text_view_data(f"({vertices[0][0]},{vertices[0][1]})",
            #                           vertices[0][0]-32, vertices[0][1], "#0000FF",
            #                           "12px Arial BOLD"))
        for wall in self.game_mode.slant_walls:
            vertices = [(wall.body.transform * v) for v in wall.box.shape.vertices]
            vertices = [self.game_mode.trnsfer_box2d_to_pygame(v) for v in vertices]
            game_info["background"].append(create_polygon_view_data("wall", vertices, WALL_COLOR))

        # add coordinate p0 p1 p2 p3
        p0 = (16, 16)
        p1 = (656, 16)
        p2 = (656, 656)
        p3 = (16, 656)

        game_info["background"].append(
            create_text_view_data(f"(0,0)",
                                  p0[0], p0[1] - 12, HELP_TXT_COLOR,
                                  "12px Arial BOLD"))
        game_info["background"].append(
            create_rect_view_data(f"P0",
                                  p0[0], p0[1], 2, 2, HELP_TXT_COLOR))

        game_info["background"].append(
            create_text_view_data(f"(200,0)",
                                  p1[0] - 16, p1[1] - 12, HELP_TXT_COLOR,
                                  "12px Arial BOLD"))
        game_info["background"].append(
            create_rect_view_data(f"P1",
                                  p1[0], p1[1], 2, 2, HELP_TXT_COLOR))

        game_info["background"].append(
            create_text_view_data(f"(200,-200)",
                                  p2[0] - 16, p2[1] + 12, HELP_TXT_COLOR,
                                  "12px Arial BOLD"))
        game_info["background"].append(
            create_rect_view_data(f"P2",
                                  p2[0], p2[1], 2, 2, HELP_TXT_COLOR))

        game_info["background"].append(
            create_text_view_data(f"(0,-200)",
                                  p3[0], p3[1] + 12, HELP_TXT_COLOR,
                                  "12px Arial BOLD"))
        game_info["background"].append(
            create_rect_view_data(f"P3",
                                  p3[0], p3[1], 2, 2, HELP_TXT_COLOR))

        for p in self.game_mode.all_points:
            point_data = p.get_progress_data()
            game_info["background"].append(point_data)
            game_info["background"].append(
                create_text_view_data(f"({p.get_info()['coordinate'][0]},{p.get_info()['coordinate'][1]})",
                                      point_data['x'] - 12, point_data['y'] + 50,
                                      HELP_TXT_COLOR,
                                      "12px Arial BOLD"))
        game_info["musics"] = [
            create_music_init_data("bgm", file_path=BGM_PATH, github_raw_url=BGM_URL),
        ]
        return game_info

    @check_game_progress
    def get_scene_progress_data(self) -> dict:
        """
        Get the position of game objects for drawing on the web
        """
        background = []
        object_list = []
        toggle = []
        toggle_with_bias = []

        # if self.is_single:
        #     # 讓鏡頭跟著
        #     game_progress["game_sys_info"] = {"view_center_coordinate": [250 - self.game_mode.car_info[0]["center"][0],
        #                                                                  240 - self.game_mode.car_info[0]["center"][1]]}
        # else:
        #     # 鏡頭固定在車子出生的位置
        #     game_progress["game_sys_info"] = {"view_center_coordinate": [250 - self.origin_car_pos[0],
        #                                                                  240 - self.origin_car_pos[1]]}
        for p in self.game_mode.all_points:
            if isinstance(p, Check_point):
                point_data = p.get_progress_data()
                object_list.append(point_data)

        # end point
        object_list.append(self.game_mode.end_point.get_progress_data())
        # rect
        # game_progress["background"].append(create_image_view_data("bg_img", 0, 0, 860, 560))
        background.append(create_image_view_data("info", 700, 0, 300, 700))
        # game_progress["toggle"].append(create_image_view_data("bg_img", 0, 0, 860, 560))
        p = self.game_mode.trnsfer_box2d_to_pygame((0, 0))
        # for x in range(TILE_LEFTTOP[0], TILE_WIDTH + TILE_LEFTTOP[0]+1, TILESIZE):
        #     game_progress["toggle_with_bias"].append(create_line_view_data("x", x, TILE_LEFTTOP[1], x, TILE_HEIGHT+TILE_LEFTTOP[1], "#8c8c8c"))
        #
        # for y in range(TILE_LEFTTOP[1], TILE_HEIGHT + TILE_LEFTTOP[1]+1, TILESIZE):
        #     game_progress["toggle_with_bias"].append(create_line_view_data("y", TILE_LEFTTOP[0], y, TILE_WIDTH+TILE_LEFTTOP[0], y, "#8c8c8c"))
        # object_list.append(create_rect_view_data("rect", p[0], p[1], 10, 10, "#356425"))
        # info
        # game_progress["toggle"].append(create_image_view_data("info", 525, 40, 327, 480))
        # car

        # text
        background.append(
            create_text_view_data(f"{self.frame_count:4d}", 800, 40, WHITE, font_style="40px Arial"))
        background.append(create_text_view_data("frames", 815, 100, WHITE, font_style="20px Arial"))
        # game_progress["toggle"].append(create_text_view_data("{0:05d} frames".format(self.frame_count), 750, 100, WHITE, font_style="36px Arial"))
        for car in self.game_mode.car_info:
            y = 170
            x = 845
            # sensor_label = 'L'
            if car["is_running"]:
                background.append(
                    create_text_view_data(f"{'L  ':<9}{car['l_sensor_value']['distance']:0>5.1f}cm",
                                          x,
                                          y + 130 * (car["id"]), YELLOW,
                                          "15px Arial BOLD")
                )
                background.append(
                    create_text_view_data(f"{'LF':<8}{car['l_t_sensor_value']['distance']:0>5.1f}cm",
                                          x,
                                          y + 16 + 130 * (car["id"]), YELLOW,
                                          "15px Arial BOLD")
                )
                background.append(
                    create_text_view_data(f"{'F  ':<9}{car['f_sensor_value']['distance']:0>5.1f}cm",
                                          x,
                                          y + 32 + 130 * (car["id"]), RED,
                                          "15px Arial BOLD")
                )
                background.append(
                    create_text_view_data(f"{'R  ':<9}{car['r_sensor_value']['distance']:0>5.1f}cm",
                                          x,
                                          y + 48 + 130 * (car["id"]), BLUE,
                                          "15px Arial BOLD")
                )
                toggle_with_bias.append(
                    create_text_view_data(f"{'RF':<8}{car['r_t_sensor_value']['distance']:0>5.1f}cm",
                                          x,
                                          y + 64 + 130 * (car["id"]), BLUE,
                                          "15px Arial BOLD")
                )

                object_list.append(
                    create_line_view_data("l_sensor", car["center"][0], car["center"][1],
                                          self.trnsfer_box2d_to_pygame(car["l_sensor_value"]["coordinate"])[0],
                                          self.trnsfer_box2d_to_pygame(car["l_sensor_value"]["coordinate"])[1],
                                          YELLOW, 3))

                object_list.append(
                    create_line_view_data("l_top_sensor", car["center"][0], car["center"][1],
                                          self.trnsfer_box2d_to_pygame(car["l_t_sensor_value"]["coordinate"])[0],
                                          self.trnsfer_box2d_to_pygame(car["l_t_sensor_value"]["coordinate"])[1],
                                          YELLOW, 3))

                object_list.append(
                    create_line_view_data("r_top_sensor", car["center"][0], car["center"][1],
                                          self.trnsfer_box2d_to_pygame(car["r_t_sensor_value"]["coordinate"])[0],
                                          self.trnsfer_box2d_to_pygame(car["r_t_sensor_value"]["coordinate"])[1],
                                          BLUE, 3))
                object_list.append(
                    create_line_view_data("r_sensor", car["center"][0], car["center"][1],
                                          self.trnsfer_box2d_to_pygame(car["r_sensor_value"]["coordinate"])[0],
                                          self.trnsfer_box2d_to_pygame(car["r_sensor_value"]["coordinate"])[1],
                                          BLUE, 3))
                object_list.append(
                    create_line_view_data("f_sensor", car["center"][0], car["center"][1],
                                          self.trnsfer_box2d_to_pygame(car["f_sensor_value"]["coordinate"])[0],
                                          self.trnsfer_box2d_to_pygame(car["f_sensor_value"]["coordinate"])[1],
                                          RED, 3))
            else:
                object_list.append(create_text_view_data("{0:4d} frames".format(car["end_frame"]),
                                                         x,
                                                         y + 32 + 130 * (car["id"]),
                                                         WHITE,
                                                         "18px Arial BOLD"))
        for car in self.game_mode.car_info:
            object_list.append(
                create_image_view_data("car_0" + str(car["id"] + 1), car["topleft"][0], car["topleft"][1],
                                       car['size'][0], car['size'][1],
                                       car["angle"])
            )
            toggle_with_bias.append(
                create_text_view_data(f"({car['coordinate'][0]},{car['coordinate'][1]})",
                                      car["topleft"][0] - 30, car["topleft"][1] + 30, CAR_COLOR[car['id']],
                                      "18px Arial BOLD"))
            """
            "x": car["coordinate"][0],
            "y": car["coordinate"][1],
            "angle": (car["angle"] * 180 / math.pi) % 360,
            """
        game_progress = create_scene_progress_data(
            frame=self.frame_count,
            background=background,
            object_list=object_list,
            foreground=[],
            toggle=toggle,
            toggle_with_bias=toggle_with_bias,
            musics=[MusicProgressSchema(music_id=f"bgm").__dict__],
            sounds=[]
        )
        return game_progress

    @check_game_result
    def get_game_result(self):
        """
        Get the game result for the web

        same_rank = {"玩家編號": str(user.car_no + 1) + "P",
        "單局排名": self.game_mode.ranked_user.index(ranking) + 1,
        "使用總幀數": user.end_frame,
        "遊戲總幀數限制":self.game_end_time,
        "使用時間百分比":round(user.end_frame/self.game_end_time,5)*100,
        "檢查點總數量":self.game_mode.check_point_num,
        "玩家通過檢查點數量": user.check_point,
        "玩家未通過檢查點數量": remain_point,
        "檢查點通過率": pass_percent,
        "檢查點未通過率": remain_percent,
        }

        """
        scene_info = self.get_scene_info
        result = self.game_mode.result
        rank = []
        for user in self.game_mode.ranked_user:
            if self.game_mode.check_point_num:

                pass_percent = round(user.check_point / self.game_mode.check_point_num, 5) * 100
                remain_point = self.game_mode.check_point_num - user.check_point
                remain_percent = 100 - pass_percent
            else:
                pass_percent = 0
                remain_point = 0
                remain_percent = 0

            same_rank = {
                "player_num": str(user.car_no + 1) + "P",
                "rank": user.rank,
                "used_frame": user.end_frame,
                "frame_limit": self.game_end_time,
                "frame_percent": round(user.end_frame / self.game_end_time * 100, 3),
                "total_checkpoints": self.game_mode.check_point_num,
                "check_points": user.check_point,
                "remain_points": remain_point,
                "pass_percent": pass_percent,
                "remain_percent": remain_percent,
                "score": user.score
            }
            rank.append(same_rank)

        return {"frame_used": scene_info["frame"],
                "status": self.game_mode.state,
                "attachment": rank,
                }

        pass

    def get_keyboard_command(self):
        """
        Get the command according to the pressed keys
        """
        if not self.isRunning():
            return {"1P": "RESET",
                    "2P": "RESET",
                    "3P": "RESET",
                    "4P": "RESET",
                    "5P": "RESET",
                    "6P": "RESET",
                    }
        key_pressed_list = pygame.key.get_pressed()
        cmd_1P = {"left_PWM": 0, "right_PWM": 0}
        cmd_2P = {"left_PWM": 0, "right_PWM": 0}
        cmd_3P = {"left_PWM": 0, "right_PWM": 0}
        cmd_4P = {"left_PWM": 0, "right_PWM": 0}
        cmd_5P = {"left_PWM": 0, "right_PWM": 0}
        cmd_6P = {"left_PWM": 0, "right_PWM": 0}

        if key_pressed_list[pygame.K_UP]:
            cmd_1P["left_PWM"] = 100
            cmd_1P["right_PWM"] = 100
        elif key_pressed_list[pygame.K_DOWN]:
            cmd_1P["left_PWM"] = -100
            cmd_1P["right_PWM"] = -100
        if key_pressed_list[pygame.K_LEFT]:
            cmd_1P["right_PWM"] += 100
        elif key_pressed_list[pygame.K_RIGHT]:
            cmd_1P["left_PWM"] += 100

        if key_pressed_list[pygame.K_w]:
            cmd_2P["left_PWM"] = 100
            cmd_2P["right_PWM"] = 100
        elif key_pressed_list[pygame.K_s]:
            cmd_2P["left_PWM"] = -100
            cmd_2P["right_PWM"] = -100
        elif key_pressed_list[pygame.K_a]:
            cmd_2P["right_PWM"] += 100
        elif key_pressed_list[pygame.K_d]:
            cmd_2P["left_PWM"] += 100

        return {"1P": cmd_1P,
                "2P": cmd_2P,
                "3P": cmd_3P,
                "4P": cmd_4P,
                "5P": cmd_5P,
                "6P": cmd_6P}

    def trnsfer_box2d_to_pygame(self, coordinate):
        '''
        :param coordinate: vertice of body of box2d object
        :return: center of pygame rect
        '''
        return (
            (coordinate[0] - self.game_mode.pygame_point[0]) * PPM,
            (self.game_mode.pygame_point[1] - coordinate[1]) * PPM)



from functools import lru_cache

import pygame

from mlgame.game.paia_game import GameResultState, GameStatus
from .env import *


class Point(pygame.sprite.Sprite):
    def __init__(self, game, coordinate):
        self.group = game.all_points
        pygame.sprite.Sprite.__init__(self, self.group)
        self.game = game
        self.x, self.y = (coordinate[0] - TILESIZE / PPM, -coordinate[1] + TILESIZE / PPM)
        # self.x, self.y = (coordinate[0] + TILESIZE / (2 * PPM), -coordinate[1] - TILESIZE / (2 * PPM))

    def get_info(self):
        return {
            "coordinate": ((self.x + 0.5) * 5, (self.y - 0.5) * 5)
        }

    def get_progress_data(self):
        asset_data = {}
        return asset_data


def trnsfer_box2d_to_pygame(coordinate):
    '''
    :param coordinate: vertice of body of box2d object
    :return: center of pygame rect
    '''
    return ((coordinate[0]) * PPM, (0 - coordinate[1]) * PPM)


class End_point(Point):
    def __init__(self, game, coordinate):
        Point.__init__(self, game, coordinate)
        self.size = TILESIZE * 3
        self.rect = pygame.Rect(self.x, self.y, self.size, self.size)
        self.rect.x, self.rect.y = trnsfer_box2d_to_pygame((self.x, self.y))

    def update(self, *args, **kwargs) -> None:
        self.detect_cars_collision()

    def detect_cars_collision(self):
        hits = pygame.sprite.spritecollide(self, self.game.cars, False)
        for hit in hits:
            if hit.is_running:
                hit.end_frame = self.game.frame
                hit.check_point += 1

                hit.is_completed = True
                self.game.eliminated_user.append(hit)  # TODO #外部注入
                if len(self.game.cars)==1:
                    self.game.state = GameResultState.PASSED
                else:
                    self.game.state = GameResultState.FINISH

                hit.is_running = False
                hit.status = GameStatus.GAME_PASS

    def get_progress_data(self):
        asset_data = {"type": "image",
                      "x": self.rect.x,
                      "y": self.rect.y,
                      "width": self.size,
                      "height": self.size,
                      "image_id": "endpoint",
                      "angle": 0}
        return asset_data


class Check_point(Point):

    def __init__(self, game, coordinate):
        Point.__init__(self, game, coordinate)
        self.size = TILESIZE * 3
        self.rect = pygame.Rect(self.x, self.y, self.size, self.size)
        self.rect.x, self.rect.y = trnsfer_box2d_to_pygame((self.x, self.y))

        self.car_has_hit = []
        self._touched = False
        self._game = game

    def update(self, *args, **kwargs) -> None:
        self.detect_cars_collision()

    def detect_cars_collision(self):
        hits = pygame.sprite.spritecollide(self, self.game.cars, False)
        if hits:
            self._touched = True
        else:
            self._touched = False

        for hit in hits:
            if hit.status and hit not in self.car_has_hit:
                hit.check_point += 1
                hit.end_frame = self.game.frame
                self.car_has_hit.append(hit)

    def get_progress_data(self):
        # asset_data = {"type": "rect",
        #               "x": self.rect.x,
        #               "y": self.rect.y,
        #               "width": 60,
        #               "height": 60,
        #               "color": RED,
        #               "angle": 0}
        asset_data = {"type": "image",
                      "x": self.rect.x,
                      "y": self.rect.y,
                      "width": self.size,
                      "height": self.size,
                      "image_id": self._image_id(self._touched),
                      "angle": 0}
        return asset_data

    @lru_cache
    def _image_id(self, touched: bool):
        if touched:
            return "checkpoint_2"
        else:
            return "checkpoint"


class Outside_point(Point):
    '''
    if car colliding these point, car will be swich to start point
    '''

    def __init__(self, game, coordinate):
        Point.__init__(self, game, coordinate)
        # self.image = pygame.Surface((TILESIZE, TILESIZE))
        # self.image.fill(BLUE)
        # self.rect = self.image.get_rect()
        self.rect = pygame.Rect(self.x, self.y, TILESIZE * 3, TILESIZE * 3)

    def update(self, *args, **kwargs) -> None:
        self.detect_cars_collision()

    def detect_cars_collision(self):
        hits = pygame.sprite.spritecollide(self, self.game.cars, False)
        for hit in hits:
            if hit.status:
                hit.body.position = (hit.x, hit.y)
                hit.body.linearVelocity = 0, 0
                pass



from .car import *
from .highway import *
from .gameMode import GameMode
from .env import *
import pygame
import random
from mlgame.game.paia_game import GameResultState, GameStatus


class PlayingMode(GameMode):
    def __init__(self, user_num: int, car_num, length, sound_controller):
        super(PlayingMode, self).__init__(user_num, car_num, length, sound_controller)
        self.is_arrive = False

    def update(self, command):
        '''update the model of src,call this fuction per frame'''
        self.count_bg()
        self.frame += 1
        self.handle_event()
        self._revise_speed() # get the velocity of user, and revise max_vel

        if self.frame > FPS:
            if self.frame > FPS * 4:
                self._creat_computercar()

            self.camera.update(self.maxVel)

            '''update sprite'''
            self.lanes.update(self.camera.position)


            for car in self.users:
                self.user_out_screen(car)
                car.update(command[str(car.car_no + 1) + "P"])

                '''是否通過終點'''
                if self._is_car_arrive_end(car):
                    self.is_arrive = True
                    self.eliminated_user.append(car)
                    self.user_distance.append(car.distance)
                    break # 任一玩家通過終點則結束遊戲

            for car in self.cars:
                '''偵測車子的狀態'''
                self._detect_car_status(car)
                '''更新車子位置'''
                car.rect.left = car.distance - self.camera.position + 520
            self.computerCars.update(self.cars)

        if self._is_game_end(self.is_arrive):
            self.rank()
            self._print_result()
            self.running = False

    def detect_collision(self):
        super(PlayingMode, self).detect_collision()
        for car in self.cars:
            self.cars.remove(car) # 如果sprite本身也在要偵測的group裡面就會被偵測到
            hits = pygame.sprite.spritecollide(car, self.cars, False)
            for hit in hits:
                if (hit.state and 0 < hit.rect.centerx < WIDTH):
                    self.sound_controller.play_hit_sound()
                hit.state = False
                car.state = False
                if car in self.users:
                    car.status = GameStatus.GAME_OVER
                if hit in self.users:
                    hit.status = GameStatus.GAME_OVER
            self.cars.add(car)

    def _print_result(self):
        '''
        依照排名順序印出玩家遊戲結果，以字典形式顯示，包含以下項目：
        'player':標示玩家代號，顯示為1P、2P等
        'distance':記錄家此局所行走的距離
        'rank':顯示完家此局排名
        :return:None
        '''
        tem = []
        for user in self.winner:
            tem.append({"player_num": str(user.car_no + 1) + "P",
                        "distance": str(round(user.distance)) + "m",
                        "used_frames": str(user.used_frame)+" frames",
                        "single_rank": self.winner.index(user) + 1
                        })
            print({"player_num": str(user.car_no + 1) + "P",
                   "distance": str(round(user.distance)) + "m",
                    "used_frames": str(user.used_frame)+" frames",
                   "single_rank":self.winner.index(user)+1
                   })
        self.winner = tem

    def _is_game_end(self, is_arrive:bool):
        '''
        判斷遊戲是否結束，遊戲結束的條件如下：
        單人模式：抵達終點，或是玩家出局(FAIL)
        多人模式：一名或以上玩家抵達終點，或市場上僅餘一名玩家
        :return:Bool
        '''
        if is_arrive:
            if self.is_single:
                self.state = GameResultState.PASSED
            else:
                self.state = GameResultState.FINISH
            return True
        if self.is_single:
            if len(self.eliminated_user) == 1:
                self.state = GameResultState.UN_PASSED
                return True
            return False
        else:
            if len(self.users) <= len(self.eliminated_user):
                self.state = GameResultState.FINISH
                return True
            else:
                return False

    def _is_car_arrive_end(self, car):
        '''
        :param car: User
        :return: Bool
        '''
        if car.distance > self.length:
            car.status = GameStatus.GAME_PASS
            return True
        return False

    def rank(self):
        for user in self.users:
            if user not in self.eliminated_user:
                user.status = GameStatus.GAME_OVER
                self.eliminated_user.append(user)
                self.user_distance.append(user.distance)
        while len(self.eliminated_user) > 0:
            for car in self.eliminated_user:
                if car.distance == min(self.user_distance):
                    self.winner.append(car)
                    self.user_distance.remove(car.distance)
                    self.eliminated_user.remove(car)
        self.winner.reverse()


from mlgame.core.communication import TransitionCommManager
from mlgame.core.model import MLGameDataType
from mlgame.executor.interface import ExecutorInterface
from mlgame.utils.logger import logger
from mlgame.view.view import PygameView


import traceback


class DisplayExecutor(ExecutorInterface):
    """current not used """
    def __init__(self, display_comm: TransitionCommManager, scene_init_data):
        # super().__init__(name="ws")
        logger.info("             display_process_init ")
        self._proc_name = "display"
        self._comm_manager = display_comm
        self._recv_data_func = self._comm_manager.recv_from_game
        self._scene_init_data = scene_init_data

    def run(self):
        self.game_view = PygameView(self._scene_init_data)
        self._comm_manager.start_recv_obj_thread()
        try:
            while (game_data := self._recv_data_func()).type != MLGameDataType.GAME_RESULT:

                if game_data.type == MLGameDataType.GAME_PROGRESS:
                    # print(game_data)
                    self.game_view.draw(game_data.data)
                    pass
        except Exception as e:
            # exception = TransitionProcessError(self._proc_name, traceback.format_exc())
            self._comm_manager.send_exception(f"exception on {self._proc_name}")
            # catch connection error
            print("except", e)
            logger.exception(traceback.format_exc())

        finally:
            print("end display process")



from mlgame.core.communication import TransitionCommManager
from mlgame.core.model import MLGameDataType, MLGameEntityWrapperSchema
from mlgame.executor.interface import ExecutorInterface
from mlgame.utils.azure import upload_data_to_azure_blob
from mlgame.utils.logger import logger


class AzureUploader(ExecutorInterface):
    _count_bias_list = [15, 15, 30, 30, 60,60,90,90,120]
    def __init__(self, pl_comm: TransitionCommManager, az_blob_url: str):
        # super().__init__(name="ws")
        self._proc_name = "azure_uploader"
        self._comm_manager = pl_comm
        self._recv_data_func = self._comm_manager.recv_from_game
        self._file_count = 0
        self._progress_data = []
        self._prev_count = 0
        self._post_count = self._get_next_count_bias()
        self._progress_count = -1
        self._az_url = str(az_blob_url)
    def run(self):
        self._comm_manager.start_recv_obj_thread()
        handlers = {
            MLGameDataType.GAME_PROGRESS: self._handle_progress,
            MLGameDataType.GAME_RESULT: self._handle_result,
            MLGameDataType.GAME_ERROR: self._handle_error,
            MLGameDataType.SYSTEM_MSG: self._handle_system_msg,
            MLGameDataType.GAME_INFO: self._handle_game_info
        }
        try:
            while game_data := self._recv_data_func():
                logger.info(f"AzureUploader received data: {game_data}")
                handler = handlers.get(game_data.type)
                if handler:
                    handler(game_data)
                else:
                    logger.warning(f"No handler found for {game_data.type}")

                if game_data.type == MLGameDataType.GAME_RESULT or game_data.type == MLGameDataType.NONE:
                    break

        except Exception as e:
            self._comm_manager.send_exception(f"Exception on {self._proc_name}")
            logger.exception(e)
        finally:
            logger.info("AzureUploader finished")
    def _handle_game_info(self, game_data: MLGameEntityWrapperSchema) -> None:
        """Handle game information data"""
        try:
            self._filename = "init.json"
            upload_data = [game_data.model_dump(mode='json')]
            upload_data_to_azure_blob(self._az_url, self._filename, upload_data)
                
        except Exception as e:
            logger.error(f"Error handling game info data: {str(e)}")
            raise

    def _handle_progress(self, game_data: MLGameEntityWrapperSchema) -> None:
        """Handle game progress data"""
        try:
            self._progress_data.append(game_data.data)
            if len(self._progress_data) > self._post_count:
                self._filename = f"{self._file_count}.json"
                upload_data = [obj.model_dump(mode='json') for obj in self._progress_data[self._prev_count:self._post_count]]
                upload_data_to_azure_blob(self._az_url, self._filename, upload_data)
                self._prev_count = self._post_count
                self._post_count += self._get_next_count_bias()
                self._file_count += 1
        except Exception as e:
            logger.exception(e)
            
    def _get_next_count_bias(self):
        if self._file_count >=len(self._count_bias_list):
            return self._count_bias_list[-1]
        return self._count_bias_list[self._file_count]
    def _handle_result(self, game_data: MLGameEntityWrapperSchema) -> None:
        """Handle game result data"""
        try:
            if self._progress_data:
                self._filename = f"{self._file_count}-end.json"
                upload_data = [obj.model_dump(mode='json') for obj in self._progress_data[self._prev_count:]]
                upload_data_to_azure_blob(self._az_url, self._filename, upload_data)
                self._prev_count = len(self._progress_data)
                self._post_count = len(self._progress_data)
                self._file_count += 1
            if game_data:
                self._filename = "result.json"
                upload_data = [game_data.model_dump(mode='json')]
                upload_data_to_azure_blob(self._az_url, self._filename, upload_data)

            logger.info(f"Game result received: {game_data.data}")
        except Exception as e:
            logger.error(f"Error handling result data: {str(e)}")
            raise

    def _handle_error(self, game_data: MLGameEntityWrapperSchema) -> None:
        """Handle game error data"""
        try:
            error_msg = game_data.data

            logger.error(
                f"Game error received - Message: {error_msg}"
            )
        except Exception as e:
            logger.error(f"Error handling error data: {str(e)}")
            raise

    def _handle_system_msg(self, game_data: MLGameEntityWrapperSchema) -> None:
        """Handle system message data"""
        try:
            message = game_data.data
            logger.info(f"System message received: {message}")
        except Exception as e:
            logger.error(f"Error handling system message: {str(e)}")
            raise



from os import path
from .version import version
PPM = 16  # pixels per meter
TARGET_FPS = 60
TIME_STEP = 1.0 / TARGET_FPS

'''width and height'''
WIDTH = 1000
HEIGHT = 700
TILE_WIDTH = 540  # 大小
TILE_HEIGHT = 540

'''tile-base'''
TILESIZE = 16
TILE_LEFTTOP = 16, 16  # pixel
GRIDWIDTH = (TILE_WIDTH + TILE_LEFTTOP[0]) / TILESIZE
GRIDHEIGHT = (TILE_HEIGHT + TILE_LEFTTOP[1]) / TILESIZE

'''sensor set trans'''
sensor_trans = ((1, 0),
                (0, 1),
                (-1, 0),
                (0, -1))

'''environment data'''
FPS = 30

'''color'''
BLACK = (0, 0, 0)
WHITE = "#F5F5F5"
RED = "#ff0000"
YELLOW = "#FFF500"
GREEN = (0, 255, 0)
GREY = "#9E9E9E"
BLUE = "#0C55FF"
LIGHT_BLUE = "#03A9F4"
BROWN = "#795548"
PINK = "#E91E63"
MEDIUMPURPLE = "#9C27B0"
WALL_COLOR = "#D9D9D9"
CAR1_COLOR = "#FFC107"
CAR2_COLOR = "#FF80AB"
CAR3_COLOR = "#00E5FF"
CAR4_COLOR = "#B2FF59"
CAR_COLOR = [CAR1_COLOR, CAR2_COLOR, CAR3_COLOR, CAR4_COLOR]
'''object size'''
car_size = (50, 40)

'''data path'''
ASSET_IMAGE_DIR = path.join(path.dirname(__file__), "../asset/image")
IMAGE_DIR = path.join(path.dirname(__file__), 'image')
SOUND_DIR = path.join(path.dirname(__file__), '../asset/sound')

'''image'''
BG_IMG = "bg_img.png"
BG_URL = 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/bg_img.png'
BG_PATH = path.join(ASSET_IMAGE_DIR, BG_IMG)
INFO_NAME = "info.png"
INFO_PATH = path.join(ASSET_IMAGE_DIR, INFO_NAME)
INFO_URL = 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/info.png'

LOGO = "logo.png"
LOGO_URL = 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/logo.png'

TMF_LOGO = "TMFlogo.png"
TMF_LOGO_URL = 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/TMFlogo.png'
ENDPOINT_IMG = "endpoint.png"
ENDPOINT_URL = f'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/{ENDPOINT_IMG}'
ENDPOINT_PATH = path.join(ASSET_IMAGE_DIR, ENDPOINT_IMG)

CHECKPOINT_IMG = "checkpoint.png"
CHECKPOINT_URL = f'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/{CHECKPOINT_IMG}'
CHECKPOINT_PATH = path.join(ASSET_IMAGE_DIR, CHECKPOINT_IMG)
CHECKPOINT2_IMG = "checkpoint_2.png"
CHECKPOINT2_URL = f'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/{CHECKPOINT2_IMG}'
CHECKPOINT2_PATH = path.join(ASSET_IMAGE_DIR, CHECKPOINT2_IMG)

CARS_NAME = ["car_01.png", "car_02.png", "car_03.png", "car_04.png", "car_05.png", "car_06.png", ]
CARS_URL = [f'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/car_01.png',
            f'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/car_02.png',
            f'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/car_03.png',
            f'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/car_04.png',
            f'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/car_05.png',
            f'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/{version}/asset/image/car_06.png'
            ]

'''map_file'''
PRACTICE_MAPS = ["level_1.json", "level_2.json", "level_3.json", "level_4.json", "level_5.json", "level_6.json"]
NORMAL_MAZE_MAPS = ["normal_map_1.json", "normal_map_2.json", "normal_map_3.json", "normal_map_3.json",
                    "normal_map_3.json", "normal_map_3.json"]
MOVE_MAZE_MAPS = ["move_map_1.json", "move_map_2.json", "move_map_3.json", "move_map_4.json", "move_map_4.json",
                  "move_map_4.json"]
BG_COLOR = "#8493B1"
MAP_FOLDER = path.join(path.dirname(__file__), "map")
HELP_TXT_COLOR = "#0000FF"

BGM_FILE_NAME = "BGM.mp3"
MUSIC_PATH = path.join(path.dirname(__file__), "..", "asset", "music")
MUSIC_URL = f"https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/maze_car/{version}/asset/music/"

BGM_PATH = path.join(MUSIC_PATH, BGM_FILE_NAME)
BGM_URL = MUSIC_URL+BGM_FILE_NAME


from mlgame.view.view_model import create_asset_init_data
from .math_function import cross_point_dot
import math

from .sensor import Sensor
import pygame
from .env import *


class Car(pygame.sprite.Sprite):
    def __init__(self, world, coordinate: tuple, car_no: int, sensor_num, angle: int):
        pygame.sprite.Sprite.__init__(self)
        self.car_no = car_no  # From 0 to 3
        self.image_num = 9
        self.image_name = f"car_0{self.car_no + 1}"
        self.collide_frame = -100
        self.collide_times = 0
        self.size = (int(PPM * 2), int(PPM * 2))  # car size
        self.is_completed = False
        self.end_frame = 0
        self.origin_image = pygame.transform.scale(
            pygame.image.load(path.join(ASSET_IMAGE_DIR, f"car_0{self.car_no + 1}.png")),
            self.size)
        self.image = self.origin_image  # after rotate
        self.rect = self.image.get_rect()
        self.is_running = True
        self.status = "GAME_ALIVE"
        self.sensor_R = {"coordinate": (0, 0), "distance": -1}
        self.sensor_L = {"coordinate": (0, 0), "distance": -1}
        self.sensor_R_T = {"coordinate": (0, 0), "distance": -1}
        self.sensor_L_T = {"coordinate": (0, 0), "distance": -1}
        self.sensor_F = {"coordinate": (0, 0), "distance": -1}
        self.sensor_B = {"coordinate": (0, 0), "distance": -1}
        self.L_PWM = 0
        self.R_PWM = 0
        self.rect.center = (0, 0)  # pygame
        # tiled -> body
        self.x, self.y = coordinate
        self.body = world.CreateDynamicBody(position=coordinate)
        self.box = self.body.CreatePolygonFixture(box=(1, 1), density=1, friction=0.1, restitution=0.3)
        self.sensor = Sensor(world, self.body, sensor_num, angle)
        self.body.angle = math.pi * angle
        self.check_point = 0
        self.explosion = False

    def update(self, commands):
        if self.explosion:
            self.image_name = f"regularExplosion0{self.image_num // 10}"
            self.image_num += 1
            if self.image_num == 60:
                self.image_num = 9
                self.explosion = False
                self.image_name = f"car_0{self.car_no + 1}"
        self.image = pygame.transform.rotate(self.origin_image, (self.body.angle * 180 / math.pi) % 360)
        self.rect = self.image.get_rect()
        if self.is_running and commands != None:
            if commands['right_PWM'] > 255:
                self.R_PWM = 255
            elif commands['right_PWM'] < -255:
                self.R_PWM = -255
            else:
                self.R_PWM = commands['right_PWM']
            if commands['left_PWM'] > 255:
                self.L_PWM = 255
            elif commands['left_PWM'] < -255:
                self.L_PWM = -255
            else:
                self.L_PWM = commands['left_PWM']
            self.left_move(self.L_PWM)
            self.right_move(self.R_PWM)
        else:
            self.body.linearVelocity = (0, 0)

    def collide(self, frame):
        if frame - self.collide_frame > 120:
            self.collide_times += 1
            self.collide_frame = frame
            self.explosion = True
            return True
        return False

    def detect_distance(self, frame, walls):
        sensor_value = self.sensor.update(frame, walls)
        self.sensor_R = sensor_value["right_value"]
        self.sensor_L = sensor_value["left_value"]
        self.sensor_R_T = sensor_value["right_top_value"]
        self.sensor_L_T = sensor_value["left_top_value"]
        self.sensor_F = sensor_value["front_value"]
        self.sensor_B = sensor_value["rear_value"]

    def left_move(self, pwm: int):
        if pwm < 0:
            self.sensor.sensor_left.linearVelocity = self.body.GetWorldVector(localVector=(0, -(abs(pwm) ** 0.5)))
        else:
            self.sensor.sensor_left.linearVelocity = self.body.GetWorldVector(localVector=(0, pwm ** 0.5))

    def right_move(self, pwm: int):
        if pwm < 0:
            self.sensor.sensor_right.linearVelocity = self.body.GetWorldVector(localVector=(0, -(abs(pwm) ** 0.5)))
        else:
            self.sensor.sensor_right.linearVelocity = self.body.GetWorldVector(localVector=(0, pwm ** 0.5))

    def get_info(self):
        self.car_info = {"id": self.car_no,
                         "status": self.status,
                         "is_running": self.is_running,
                         "size": self.size,  # pygame
                         "topleft": self.rect.topleft,  # pygame
                         "center": self.rect.center,
                         "coordinate": (
                             round((self.body.position[0]) * 5, 2),
                             round((self.body.position[1]) * 5, 2)),
                         "angle": self.body.angle,  # Box2D
                         "r_sensor_value": self.sensor_R,
                         "l_sensor_value": self.sensor_L,
                         "r_t_sensor_value": self.sensor_R_T,
                         "l_t_sensor_value": self.sensor_L_T,
                         "f_sensor_value": self.sensor_F,
                         "b_sensor_value": self.sensor_B,
                         "L_PWM": self.L_PWM,
                         "R_PWM": self.R_PWM,
                         "end_frame": self.end_frame,
                         "image": self.image_name,
                         "crash_times": self.collide_times,
                         "check_point": self.check_point,
                         }
        return self.car_info



import os
from pathlib import Path
from loguru import logger

# Try to load environment variables from .env file in the tests directory
try:
    from dotenv import load_dotenv
    # Get the path to the tests directory
    TESTS_DIR = os.path.dirname(os.path.abspath(__file__))
    # Load from .env file in tests directory
    load_dotenv(os.path.join(TESTS_DIR, '.env'))
except ImportError:
    # python-dotenv is not installed, continue without it
    pass

# Define environment variables with defaults for tests
# Base project path (default uses current value for backward compatibility)
BASE_PATH = os.environ.get('MLGAME_BASE_PATH', os.path.join(os.path.dirname(__file__),"..",".."))
# Game and AI client paths 
GAMES_PATH = os.environ.get('MLGAME_GAMES_PATH', os.path.join(BASE_PATH, 'games'))
AI_CLIENTS_PATH = os.environ.get('MLGAME_AI_CLIENTS_PATH', os.path.join(BASE_PATH, 'ai_clients'))
OUTPUT_PATH = os.environ.get('MLGAME_OUTPUT_PATH', os.path.join(BASE_PATH, 'var'))

# Azure Storage URLs - using placeholder values as defaults
# These should be overridden with actual values in .env file or environment variables
AZURE_CONTAINER_URL = os.environ.get('MLGAME_AZURE_CONTAINER_URL', 
                                     'https://example.blob.core.windows.net/container?sv=xxxx&sig=xxxx')
AZURE_BLOB_URL = os.environ.get('MLGAME_AZURE_BLOB_URL',
                               'https://example.blob.core.windows.net/container/path?sv=xxxx&sig=xxxx')

# Helper function to create full paths
def get_path(*parts):
    return str(Path(os.path.join(*parts)))



import random

import pygame.event
import pygame.event
from mlgame.game.paia_game import GameResultState, GameStatus
from mlgame.utils.enum import get_ai_name
from mlgame.view.view_model import create_asset_init_data, create_text_view_data, \
    create_rect_view_data, create_line_view_data
from mlgame.view.view_model import create_image_view_data

from src.game_module.SoundController import create_sounds_data, create_bgm_data, SoundController
from src.game_module.TiledMap import create_construction, TiledMap
from .Bullet import Bullet
from .Player import Player
from .Gun import Gun
from .Station import Station
from .Wall import Wall
from .collide_hit_rect import *
from .env import *
from .game_module.fuctions import set_topleft, add_score, set_shoot
from .GenerateMap import MapGenerator


class TeamBattleMode:
    def __init__(self, green_team_num: int, blue_team_num: int, is_manual: bool, frame_limit: int, sound_path: str, play_rect_area: pygame.Rect):
        # init game
        pygame.init()
        self.sound_path = sound_path
        self.green_team_num = green_team_num
        self.blue_team_num = blue_team_num if (6 - (green_team_num + blue_team_num)) >= 0 else (6 - green_team_num)
        self.map_name = f"map_{green_team_num}_v_{self.blue_team_num}.tmx" if not IS_DEBUG else f"test_map_{green_team_num}_v_{self.blue_team_num}.tmx"
        self.map_path = path.join(MAP_DIR, self.map_name)
        self.map_generator = MapGenerator(self.green_team_num, self.blue_team_num, 40, 24)
        self.tileSize = self.map_generator.getTileSize()
        self.size_multiplier = self.tileSize / 50
        self.map_generator.generate_map()
        self.map = TiledMap(self.map_path)
        self.scene_width, self.scene_height = self.map_generator.getScreeenSize()
        self.width_center = self.scene_width // 2
        self.height_center = self.scene_height // 2
        self.play_rect_area = play_rect_area
        self.play_rect_area.width = self.scene_width
        self.used_frame = 0
        self.state = GameResultState.FAIL
        self.status = GameStatus.GAME_ALIVE
        self.sound_controller = SoundController(sound_path, self.get_sound_data())
        self.sound_controller.play_music(self.get_bgm_data())
        self.frame_limit = frame_limit
        self.is_manual = is_manual
        self.obj_rect_list = []
        self.team_green_score = 0
        self.team_blue_score = 0

        # control variables
        self.is_invincible = False
        self.is_through_wall = False
        # initialize sprites group
        self.all_sprites = pygame.sprite.Group()
        self.players_a = pygame.sprite.Group()
        self.players_b = pygame.sprite.Group()
        self.all_players = pygame.sprite.Group()
        self.guns = pygame.sprite.Group()
        self.walls = pygame.sprite.Group()
        self.bullets = pygame.sprite.Group()
        self.bullet_stations = pygame.sprite.Group()
        self.oil_stations = pygame.sprite.Group()
        # init players
        act_cd = 0
        if self.is_manual:
            act_cd = 10
        # init obj data
        self.map.add_init_obj_data(PLAYER_1_IMG_NO, Player, act_cd=act_cd, play_rect_area=self.play_rect_area)
        self.map.add_init_obj_data(PLAYER_2_IMG_NO, Player, act_cd=act_cd, play_rect_area=self.play_rect_area)
        self.map.add_init_obj_data(WALL_IMG_NO, Wall, margin=8, spacing=8)
        self.map.add_init_obj_data(BULLET_STATION_IMG_NO, Station, spawn_cd=30, margin=2, spacing=2, capacity=5, quadrant=1)
        self.map.add_init_obj_data(OIL_STATION_IMG_NO, Station, spawn_cd=30, margin=2, spacing=2, capacity=30, quadrant=1)
        # create obj
        all_obj = self.map.create_init_obj_dict()
        # init players
        self.players_a.add(all_obj[PLAYER_1_IMG_NO])
        self.players_b.add(all_obj[PLAYER_2_IMG_NO])
        no = 1
        for player in self.players_a:
            player.no = no
            no += 1
        for player in self.players_b:
            player.no = no
            no += 1
        self.all_players.add(*self.players_a, *self.players_b)
        self.guns.add(*[player.gun for player in self.all_players])
        self.all_sprites.add(*self.players_a, *self.players_b)
        self.all_sprites.add(*[player.gun for player in self.all_players])
        # init walls
        self.walls.add(all_obj[WALL_IMG_NO])
        self.all_sprites.add(*self.walls)
        # init bullet stations
        self.bullet_stations.add(all_obj[BULLET_STATION_IMG_NO])
        self.all_sprites.add(*self.bullet_stations)
        # init oil stations
        self.oil_stations.add(all_obj[OIL_STATION_IMG_NO])
        self.all_sprites.add(*self.oil_stations)
        # init pos list
        self.all_pos_list = self.map.all_pos_list
        self.empty_quadrant_pos_dict = self.map.empty_quadrant_pos_dict
        self.background = []
        for pos in self.all_pos_list:
            no = random.randrange(3)
            self.background.append(
                create_image_view_data(f"floor_{no}", pos[0], pos[1], self.tileSize, self.tileSize, 0))
        self.obj_list = [self.oil_stations, self.bullet_stations, self.bullets, self.all_players, self.guns, self.walls]
        self.background.append(create_image_view_data("border", 0, -50, self.scene_width, WINDOW_HEIGHT, 0))
        # self.background.append(create_image_view_data("border", 0, -self.tileSize, self.scene_width, self.scene_height, 0))

    def update(self, command: dict):
        # refactor
        self.team_green_score = sum([player.score for player in self.players_a if isinstance(player, Player)])
        self.team_blue_score = sum([player.score for player in self.players_b if isinstance(player, Player)])
        self.used_frame += 1
        self.check_collisions()
        self.walls.update()
        self.create_bullet(self.all_players)
        self.bullets.update()
        self.bullet_stations.update()
        self.oil_stations.update()
        self.all_players.update(command)
        self.get_player_end()
        if self.used_frame >= self.frame_limit:
            self.get_game_end()

    def reset(self):
        # reset init game
        self.__init__(self.green_team_num, self.blue_team_num, self.is_manual, self.frame_limit, self.sound_path, self.play_rect_area)
        # reset player pos
        self.change_player_pos()

    def get_player_end(self):
        is_alive_team_green = False
        is_alive_team_blue = False
        for player in self.all_players:
            if isinstance(player, Player) and player.is_alive:
                if player.no > self.green_team_num and not is_alive_team_blue:
                    is_alive_team_blue = True
                elif player.no <= self.green_team_num:
                    is_alive_team_green = True

        if not is_alive_team_blue:
            self.set_result(GameResultState.FINISH, "GREEN_TEAM_WIN")
        elif not is_alive_team_green:
            self.set_result(GameResultState.FINISH, "BLUE_TEAM_WIN")

    def get_game_end(self):
        if self.team_green_score > self.team_blue_score:
            self.set_result(GameResultState.FINISH, "GREEN_TEAM_WIN")
        elif self.team_green_score < self.team_blue_score:
            self.set_result(GameResultState.FINISH, "BLUE_TEAM_WIN")
        else:
            self.set_result(GameResultState.FINISH, GameStatus.GAME_DRAW)

    def set_result(self, state: str, status: str):
        self.state = state
        self.status = status

    def get_player_result(self) -> list:
        """Define the end of game will return the player's info for user"""
        res = []
        for player in self.all_players:
            if isinstance(player, Player):
                if player.no > self.green_team_num:
                    team_id = "blue"
                else:
                    team_id = "green"
                get_res = player.get_info_to_game_result()
                get_res["no"] = f"{team_id}_{player.no}P"
                get_res["state"] = self.state
                get_res["status"] = self.status
                get_res["used_frame"] = self.used_frame
                res.append(get_res)
        return res

    def check_collisions(self):
        if not self.is_through_wall:
            collide_with_walls(self.all_players, self.walls)
        if not self.is_invincible:
            player_score_data = collide_with_bullets(self.all_players, self.bullets, self.green_team_num)
            for player, score in player_score_data.items():
                self.add_player_score(player, score)
            # TODO refactor stations

            supply_stations = []

            # Check collision between player and supply stations
            supply_stations.extend(collide_with_supply_stations(self.all_players, self.bullet_stations))
            supply_stations.extend(collide_with_supply_stations(self.all_players, self.oil_stations))

            # Check collision between bullet and supply stations
            supply_stations.extend(collide_with_supply_stations(self.bullets, self.bullet_stations))
            supply_stations.extend(collide_with_supply_stations(self.bullets, self.oil_stations))

            # Update stations position
            self.change_obj_pos(supply_stations)

        player_score_data = collide_with_bullets(self.walls, self.bullets)
        for player, score in player_score_data.items():
            self.add_player_score(player, score)

    def change_player_pos(self):
        for player in self.all_players:
            quadrant = player.quadrant
            self.empty_quadrant_pos_dict[quadrant].append(player.rect.topleft)
            if quadrant == 2 or quadrant == 3:
                player.quadrant = random.choice([2, 3])
            else:
                player.quadrant = random.choice([1, 4])
            quadrant = player.quadrant
            new_pos = self.empty_quadrant_pos_dict[quadrant].pop(
                random.randrange(len(self.empty_quadrant_pos_dict[quadrant])))
            set_topleft(player, new_pos)
            set_topleft(player.gun, new_pos)

    # TODO move method to Station
    def change_obj_pos(self, objs=None):
        if objs is None:
            return
        for obj in objs:
            quadrant = obj.quadrant
            self.empty_quadrant_pos_dict[quadrant].append(obj.rect.topleft)
            if quadrant == 2 or quadrant == 3:
                obj.quadrant = random.choice([2, 3])
            else:
                obj.quadrant = random.choice([1, 4])
            quadrant = obj.quadrant
            new_pos = self.empty_quadrant_pos_dict[quadrant].pop(
                random.randrange(len(self.empty_quadrant_pos_dict[quadrant])))
            set_topleft(obj, new_pos)

    def create_bullet(self, sprites: pygame.sprite.Group):
        for sprite in sprites:
            if not sprite.is_shoot:
                continue            
            self.sound_controller.play_sound("shoot", 0.03, -1)

            init_data = create_construction(sprite.id, sprite.no, sprite.rect.center, (BULLET_SIZE[0], BULLET_SIZE[1]))
            bullet = Bullet(init_data, rot=sprite.gun.get_rot(), margin=2, spacing=2, bullet_speed=BULLET_SPEED, bullet_travel_distance=BULLET_TRAVEL_DISTANCE

                            , play_rect_area=self.play_rect_area)
            self.bullets.add(bullet)
            self.all_sprites.add(bullet)
            set_shoot(sprite, False)

    def get_init_image_data(self):
        init_image_data = []
        for i in range(3):
            init_image_data.append(create_asset_init_data(f"floor_{i}", 50, 50
                                                          , path.join(IMAGE_DIR, f"grass_{i}.png"),
                                                          f"https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/grass_{i}.png"))
        for i in range(15):
            init_image_data.append(create_asset_init_data(f"hourglass_{i}", 42, 42
                                                          , path.join(IMAGE_DIR, f"hourglass_{i}.png"),
                                                          f"https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/hourglass_{i}.png"))
        for station in self.bullet_stations:
            if isinstance(station, Station):
                for data in station.get_obj_init_data():
                    init_image_data.append(data)
                break
        for wall in self.walls:
            if isinstance(wall, Wall):
                for data in wall.get_obj_init_data():
                    init_image_data.append(data)
                break
        img_id = ["team_a_bullet", "team_b_bullet"]
        for id in img_id:
            img_url = f"https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/{id}.svg"
            bullet_image_init_data = create_asset_init_data(id, BULLET_SIZE[0], BULLET_SIZE[1],
                                                            path.join(IMAGE_DIR, f"{id}.png"), img_url)
            init_image_data.append(bullet_image_init_data)
        border_image_init_data = create_asset_init_data("border", self.scene_width, WINDOW_HEIGHT,
                                                        path.join(IMAGE_DIR, "border.png"),
                                                        f"https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/border.svg")
        init_image_data.append(border_image_init_data)
        for player in self.all_players:
            if isinstance(player, Player):
                data = player.get_obj_init_data()
                init_image_data.append(data[0])
                init_image_data.append(data[1])
                break
        for gun in self.guns:
            if isinstance(gun, Gun):
                data = gun.get_obj_init_data()
                init_image_data.append(data[0])
                init_image_data.append(data[1])
                break
        for i in range(1, 4):
            team_a_lives = "team_a_lives"
            team_a_lives_image_init_data = create_asset_init_data(f"{team_a_lives}_{i}", LIVES_SIZE[0], LIVES_SIZE[1], path.join(IMAGE_DIR, f"{team_a_lives}_{i}.png"),
                                                                  f"https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/{team_a_lives}_{i}.svg")
            init_image_data.append(team_a_lives_image_init_data)
            team_b_lives = "team_b_lives"
            team_b_lives_image_init_data = create_asset_init_data(f"{team_b_lives}_{i}", LIVES_SIZE[0], LIVES_SIZE[1], path.join(IMAGE_DIR, f"{team_b_lives}_{i}.png"),
                                                                  f"https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/{team_b_lives}_{i}.svg")
            init_image_data.append(team_b_lives_image_init_data)
        return init_image_data

    def get_toggle_progress_data(self):
        toggle_data = []
        hourglass_index = 0
        hourglass_size = self.tileSize
        if self.is_manual:
            hourglass_index = self.used_frame // 10 % 15
        toggle_data.append(
            create_image_view_data(image_id=f"hourglass_{hourglass_index}", x=0, y=2, width=hourglass_size, height=hourglass_size, angle=0))
        x = 28
        y = 8
        for frame in range((self.frame_limit - self.used_frame) // int((30 * 2))):
            toggle_data.append(create_rect_view_data("frame", x, y, 3, 15, RED))
            x += 3.5
        toggle_data.append(create_text_view_data(f"Frame: {self.frame_limit - self.used_frame}",
                                                 self.scene_width-165, 8, RED,
                                                 "24px Arial BOLD"))
        x = 28
        y = 25
        for score in range(min(self.team_green_score, self.team_blue_score)):
            toggle_data.append(create_rect_view_data(name="score", x=x, y=y, width=1, height=10, color=ORANGE))
            x += 1.5
            if x > self.width_center:
                if y == 25:
                    y = 36
                else:
                    y = 25
                x = 28
        for score in range(abs(self.team_green_score - self. team_blue_score)):
            if self.team_green_score > self.team_blue_score:
                toggle_data.append(create_rect_view_data("score", x, y, 1, 10, DARKGREEN))
            else:
                toggle_data.append(create_rect_view_data("score", x, y, 1, 10, BLUE))
            x += 1.5
            if x > self.width_center:
                if y == 32:
                    y = 44
                else:
                    y = 32
                x = 24
        # 1P
        x = self.scene_width - 125
        y = self.scene_height - 40
        toggle_data.append(create_text_view_data(f"Score: {self.team_green_score}", x, y, DARKGREEN, "24px Arial BOLD"))
        # 2P
        x = 5
        toggle_data.append(create_text_view_data(f"Score: {self.team_blue_score}", x, y, BLUE, "24px Arial BOLD"))
        for player in self.all_players:
            if isinstance(player, Player) and player.is_alive:
                # lives
                team_id = "team_a_lives" if player.id == 1 else "team_b_lives"
                color = DARKGREEN if player.id == 1 else BLUE
                x = player.play_rect_area.midbottom[0] + 7 + (player.no - 1) * 60 if player.id == 1 \
                    else player.play_rect_area.midbottom[0] - (player.no - self.green_team_num) * 60
                y = player.play_rect_area.height + 73
                toggle_data.append(
                    create_text_view_data(f"{player.no}P", x - 5, y - 25, color, "22px Arial BOLD"))
                for live in range(1, player.lives+1):
                    toggle_data.append(create_image_view_data(f"{team_id}_{live}", x, y, LIVES_SIZE[0], LIVES_SIZE[1])) 
                    x += 10
                    y -= 10
        return toggle_data

    def get_toggle_with_bias_data(self):
        toggle_with_bias_data = []
        color = WHITE
        for player in self.all_players:
            if isinstance(player, Player) and player.is_alive:
                # number
                if player.no > self.green_team_num:
                    color = WHITE
                x = player.rect.x
                y = player.rect.y - 18
                toggle_with_bias_data.append(create_text_view_data(f"{player.no}P", x, y, color, "16px Arial BOLD"))
                team_id = "team_a"
                if player.no > self.green_team_num:
                    team_id = "team_b"
                # oil
                y = player.rect.bottom
                multiplier = self.tileSize / 50
                x = player.rect.x - 25 * (1 - multiplier)
                # toggle_with_bias_data.append(create_rect_view_data(f"{team_id}_oil", x, y, int(player.oil*0.5)*multiplier, 8*multiplier, ORANGE))
                toggle_with_bias_data.append(create_rect_view_data(f"{team_id}_oil", x, y, int(player.oil*0.5), 8, ORANGE))

                # power
                y = player.rect.bottom + 10
                # x = player.rect.midbottom - 4
                for power in range(player.power):
                    toggle_with_bias_data.append(create_rect_view_data(f"{team_id}_power", x+1, y, 3, 8, BLUE))
                    # toggle_with_bias_data.append(create_rect_view_data(f"{team_id}_power", x+1, y, 3*multiplier, 8*multiplier, BLUE))
                    x += 5

        return toggle_with_bias_data

    def get_ai_data_to_player(self):
        to_player_data = {}
        num = 0
        competitor_info = {1: [player.get_data_from_obj_to_game() for player in self.players_a if isinstance(player, Player)]
                           , 2: [player.get_data_from_obj_to_game() for player in self.players_b if isinstance(player, Player)]
                           }
        walls_info = [wall.get_data_from_obj_to_game() for wall in self.walls if isinstance(wall, Wall)]
        bullet_stations_info = [bullst_station.get_data_from_obj_to_game() for bullst_station in self.bullet_stations if
                                isinstance(bullst_station, Station)]
        oil_stations_info = [oil_station.get_data_from_obj_to_game() for oil_station in self.oil_stations if
                             isinstance(oil_station, Station)]
        bullets_info = [bullet.get_data_from_obj_to_game() for bullet in self.bullets if
                             isinstance(bullet, Bullet)]
        for player in self.players_a:
            if isinstance(player, Player):
                to_game_data = player.get_data_from_obj_to_game()
                to_game_data["used_frame"] = self.used_frame
                to_game_data["status"] = self.status
                to_game_data["teammate_info"] = competitor_info[1]
                to_game_data["competitor_info"] = competitor_info[2]
                to_game_data["walls_info"] = walls_info
                to_game_data["bullets_info"] = bullets_info
                to_game_data["bullet_stations_info"] = bullet_stations_info
                to_game_data["oil_stations_info"] = oil_stations_info
                to_game_data["size_multiplier"] = self.size_multiplier
                to_player_data[get_ai_name(num)] = to_game_data
                num += 1
        for player in self.players_b:
            if isinstance(player, Player):
                to_game_data = player.get_data_from_obj_to_game()
                to_game_data["used_frame"] = self.used_frame
                to_game_data["status"] = self.status
                to_game_data["teammate_info"] = competitor_info[2]
                to_game_data["competitor_info"] = competitor_info[1]
                to_game_data["walls_info"] = walls_info
                to_game_data["bullets_info"] = bullets_info
                to_game_data["bullet_stations_info"] = bullet_stations_info
                to_game_data["oil_stations_info"] = oil_stations_info
                to_game_data["size_multiplier"] = self.size_multiplier
                to_player_data[get_ai_name(num)] = to_game_data
                num += 1

        return to_player_data

    def get_bgm_data(self):
        return create_bgm_data("BGM.ogg", 0.1)

    def get_sound_data(self):
        return [create_sounds_data("shoot", "shoot.wav")
            , create_sounds_data("touch", "touch.wav")]

    def add_player_score(self, player_no: int, score: int):
        if not player_no or not score:
            return
        for player in self.all_players:
            if isinstance(player, Player) and player_no == player.no and player.lives >= 0:
                add_score(player, score)

    def debugging(self, is_debug: bool):
        self.obj_rect_list = []
        if not is_debug:
            return
        play_rect_area_points = [self.play_rect_area.topleft, self.play_rect_area.topright
                                 , self.play_rect_area.bottomright, self.play_rect_area.bottomleft
                                 , self.play_rect_area.topleft]

        for sprite in self.all_sprites:
            if isinstance(sprite, pygame.sprite.Sprite):
                top_left = sprite.rect.topleft
                points = [top_left, sprite.rect.topright, sprite.rect.bottomright
                    , sprite.rect.bottomleft, top_left]
                for index in range(len(points) - 1):
                    self.obj_rect_list.append(create_line_view_data("rect", *points[index], *points[index + 1], RED))
                    self.obj_rect_list.append(create_line_view_data("play_rect_area", *play_rect_area_points[index]
                                                                    , *play_rect_area_points[index + 1], RED))



import random


class Scene:
    def __init__(self, width: int, height: int, color: str = "#000000", bias_x=0, bias_y=0):
        """
        This is a value object
        :param width:
        :param height:
        :param color:
        """
        self.width = width
        self.height = height
        self.color = color
        self.bias_x = bias_x
        self.bias_y = bias_y


def create_asset_init_data(image_id: str, width: int, height: int, file_path: str, github_raw_url: str):
    # assert file_path is valid
    return {
        "type": "image",
        "image_id": image_id,
        "width": width,
        "height": height,
        "file_path": file_path,
        "url": github_raw_url
    }



def create_scene_view_data(width: int, height: int, color: str = "#000000", bias_x=0, bias_y=0):
    return {
        "width": width,
        "height": height,
        "color": color,
        "bias_x": bias_x,
        "bias_y": bias_y
    }


def create_scene_progress_data(frame: int = 0, background=None, object_list=None,
                               toggle=None, toggle_with_bias=None, foreground=None, user_info=None,
                               game_sys_info=None, musics=None, sounds=None, ):
    if sounds is None:
        sounds = []
    if musics is None:
        musics = []
    if background is None:
        background = []
    if object_list is None:
        object_list = []
    if toggle is None:
        toggle = []
    if toggle_with_bias is None:
        toggle_with_bias = []
    if foreground is None:
        foreground = []
    if user_info is None:
        user_info = []
    if game_sys_info is None:
        game_sys_info = {}


    return {
        # background view data will be draw first
        "frame": frame,
        "background": background,
        # game object view data will be draw on screen by order , and it could be shifted by WASD
        "object_list": object_list,
        "toggle_with_bias": toggle_with_bias,
        "toggle": toggle,
        "foreground": foreground,
        # other information to display on web
        "user_info": user_info,
        # other information to display on web
        "game_sys_info": game_sys_info,
        "musics":musics,
        "sounds":sounds

    }


def create_image_view_data(image_id, x, y, width, height, angle=0):
    """
    這是一個用來繪製圖片的資料格式，
    "type"表示不同的類型
    "x" "y" 表示物體左上角的座標
    "width" "height"表示其大小
    "image_id"表示其圖片的識別號，需在
    "angle"表示其順時針旋轉的角度
    """
    return {
        "type": "image",
        "x": x,
        "y": y,
        "width": width,
        "height": height,
        "image_id": image_id,
        "angle": angle
    }


def create_rect_view_data(name: str, x: int, y: int, width: int, height: int, color: str, angle: int = 0):
    # TODO angle
    """
    這是一個用來繪製矩形的資料格式，
    "type"表示不同的類型
    "name"用來描述這個物件
    "x""y"表示其位置，位置表示物體左上角的座標
    "size"表示其大小
    "image"表示其圖片
    "angle"表示其順時針旋轉的角度
    "color"以字串表示
    :return:
    """
    return {
        "type": "rect",
        "name": name,
        "x": x,
        "y": y,
        "angle": angle,
        "width": width,
        "height": height,
        "color": color
    }


def create_line_view_data(name: str, x1: int, y1: int, x2: int, y2: int, color: str, width: int = 2):
    """
    這是一個用來繪製直線的資料格式，
    "x1","y1"表示起點位置，位置表示物體左上角的座標
    "x2","y2"表示終點位置，位置表示物體左上角的座標
    "color"以字串表示
    "width" 表示寬度
    :return:
    """
    return {
        "type": "line",
        "name": name,
        "x1": x1,
        "y1": y1,
        "x2": x2,
        "y2": y2,
        "width": width,
        "color": color
    }


def create_polygon_view_data(name: str, points: list, color: str):
    """
    這是一個用來繪製多邊形的資料格式，
    points欄位至少三個 # [[100,101],[52.1,31.3],[53.1,12.3]]
    :return:dict
    """
    assert len(points) >= 3
    vertices = []
    for p in points:
        vertices.append({"x": p[0], "y": p[1]})
    return {
        "type": "polygon",
        # "type": "aapolygon",

        "name": name,
        "color": color,
        "points": vertices
    }

def create_aapolygon_view_data(name: str, points: list, color: str):
    """
    這是一個用來繪製多邊形的資料格式，
    points欄位至少三個 # [[100,101],[52.1,31.3],[53.1,12.3]]
    :return:dict
    """
    assert len(points) >= 3
    vertices = []
    for p in points:
        vertices.append({"x": p[0], "y": p[1]})
    return {
        "type": "aapolygon",
        "name": name,
        "color": color,
        "points": vertices
    }


def create_text_view_data(content: str, x: int, y: int, color: str, font_style="24px Arial"):
    return {
        "type": "text",
        "content": content,
        "color": color,
        "x": x,
        "y": y,
        "font-style": font_style
    }


def get_scene_init_sample_data() -> dict:
    """
    :rtype: dict
    :return:  遊戲場景初始化的資料
    """

    scene = Scene(800, 600)
    assets = [
        {
            "type": "image",
            "image_id": 'car_01',
            "width": 50,
            "height": 50,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/car_01.png'
        }, {
            "type": "image",
            "image_id": 'car_02',
            "width": 50,
            "height": 50,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/car_02.png'
        }
    ]

    return {
        "scene": scene.__dict__,
        "assets": assets,
        # "audios": {}
    }


def get_dummy_progress_data():
    background = create_image_view_data("background", 0, 0, 800, 600)
    score_text = create_text_view_data("Score = 1", 650, 50, "#FF0000")
    rect = create_rect_view_data("dummy_rect", 200, 300, 100, 200, "#FFFAAA", 30)
    line = create_line_view_data("dummy_line", 10, 30, 100, 300, "#AAAFFF", 5)
    points = gen_points(5)
    polygon = create_polygon_view_data("dummy_polygon", points, "#FFAAFF")

    scene_progress = {
        # background view data will be draw first
        "frame":1,
        "background": [
            background,

        ],
        # game object view data will be draw on screen by order , and it could be shifted by WASD
        "object_list": [
            rect,
            line
        ],
        "toggle": [
            polygon
        ],
        "toggle_with_bias": [
            polygon
        ],
        "foreground": [
            score_text
        ],
        "game_sys_info": {},
        "musics":[],
        "sounds":[],
    }
    return scene_progress


def gen_points(point_num: int = 4) -> list:
    """
    points should be [x,y] ex [100.3,300.231]
    """
    result = []
    for i in range(point_num):
        result.append([random.randint(0, 100), random.randint(0, 100)])
    return result



import math

def cross_point(dot1, vec1, dot2, vec2):
    '''
    define line A and line B, write a function which can return the point two lines cross.
    dot_1 = (x1, y1)
    dot_2 = (x2, y2)
    vec_1 = (vx1, vy1)
    vec_2 = (vx2, vy2)
    line1 = [(x1, y1), (x1 + vx1, y1 + vy1)], line2 = [(x2, y2), (x2 + vx2, y2 + vy2)]
    '''
    x1 = dot1[0]
    y1 = dot1[1]
    x2 = dot2[0]
    y2 = dot2[1]
    vx1 = vec1[0]
    vy1 = vec1[1]
    vx2 = vec2[0]
    vy2 = vec2[1]
    if vx1 == 0:  # 如果斜率為0
        k1 = None
        b1 = 0
    else:
        k1 = vy1 * 1.0 / vx1
        b1 = (y1 + vy1) * 1.0 - (x1 + vx1) * k1 * 1.0
    if vx2 == 0:
        k2 = None
        b2 = 0
    else:
        k2 = vy2 * 1.0 / vx2
        b2 = (y2 + vy2) * 1.0 - (x2 + vx2) * k2 * 1.0

    if k1 == k2:
        return None
    elif k1 == None:  # 如果Line1斜率不存在，則取Line1上的點帶入Line2的公式
        x = x1 + vx1
        k1 = k2
        b1 = b2
    elif k2 == None:
        x = x2 + vx2
    else:
        x = (b2 - b1) * 1.0 / (k1 - k2)
    y = k1 * x * 1.0 + b1 * 1.0
    return (x, y)

def cross_point_dot(dot1, vec1, dot2, dot3):
    '''
    this function is same as above. But in this case, one of lines has starting point and ending point.
    If the point two line cross out of the line, function should return None.
    '''
    x2 = dot2[0]
    y2 = dot2[1]
    x3 = dot3[0]
    y3 = dot3[1]
    if vec1[0] == 0:
        vec1[0] = 0.1
    if x3 - x2 == 0:
        x3 += 0.1
    p = cross_point(dot1, vec1, dot2, (x3 - x2, y3 - y2))

    if p:
        if x2 -0.1 <= p[0] <= x3 + 0.1 or x3 - 0.1 <= p[0] <= x2 + 0.1:
            if y2 - 0.1 <= p[1] <= y3 + 0.1 or y3 - 0.1 <= p[1] <= y2 + 0.1:
                return p
            else:
                return None
    else:
        return None

if __name__ == '__main__':
    # p = cross_point((6.55931, -18.08653), (0.72216, 1.87358), (1.0, -12.0), (26.0, 0.0))
    # print(p)
    p = cross_point_dot((6.55931, -18.08654), (0.72215, 1.87358), (1.0, -12.0), (27.0, -12.0))
    print(p)



"""
The helper functions for physics
"""

from pygame import Rect
from pygame.sprite import Sprite
from pygame.math import Vector2

def collide_or_contact(sprite_a: Sprite, sprite_b: Sprite) -> bool:
    """
    Check if two sprites are colliding or contacting
    """
    rect_a = sprite_a.rect
    rect_b = sprite_b.rect

    if (rect_a.left <= rect_b.right and
        rect_a.right >= rect_b.left and
        rect_a.top <= rect_b.bottom and
        rect_a.bottom >= rect_b.top):
        return True
    return False

def moving_collide_or_contact(moving_sprite: Sprite, sprite: Sprite) -> bool:
    """
    Check if the moving sprite collides or contacts another sprite.

    @param moving_sprite The sprite that moves in the scene.
           It must contain `rect` and `last_pos` attributes, which both are `pygame.Rect`.
    @param sprite The sprite that will be collided or contacted by `moving_sprite`.
           It must contain `rect` attribute, which is also `pygame.Rect`.
    """
    # Generate the routine of 4 corners of the moving sprite
    move_rect = moving_sprite.rect
    move_last_pos = moving_sprite.last_pos
    routines = (
        (Vector2(move_last_pos.topleft), Vector2(move_rect.topleft)),
        (Vector2(move_last_pos.topright), Vector2(move_rect.topright)),
        (Vector2(move_last_pos.bottomleft), Vector2(move_rect.bottomleft)),
        (Vector2(move_last_pos.bottomright), Vector2(move_rect.bottomright))
    )

    # Check any of routines collides the rect
    ## Take the bottom and right into account when using the API of pygame
    rect_expanded = sprite.rect.inflate(1, 1)
    for routine in routines:
        # Exclude the case that the `moving_sprite` goes from the surface of `sprite`
        if (not rect_expanded.collidepoint(routine[0]) and
            rect_collideline(sprite.rect, routine)):
            return True

    return False

def line_intersect(line_a, line_b) -> bool:
    """
    Check if two line segments intersect

    @param line_a A tuple (Vector2, Vector2) representing both end points
           of line segment
    @param line_b Same as `line_a`
    """
    # line_a and line_b have the same end point
    if (line_a[0] == line_b[0] or
        line_a[1] == line_b[0] or
        line_a[0] == line_b[1] or
        line_a[1] == line_b[1]):
        return True

    # Set line_a to (u0, u0 + v0) and p0 = u0 + s * v0, and
    # set line_b to (u1, u1 + v1) and p1 = u1 + t * v1,
    # where u, v, p are vectors and s, t is in [0, 1].
    # If line_a and line_b intersects, then p0 = p1
    # -> u0 - u1 = -s * v0 + t * v1
    # -> | u0.x - u1.x |   | v0.x  v1.x | |-s |
    #    | u0.y - u1.y | = | v0.y  v1.y | | t |
    #
    # If left-hand vector is a zero vector, then two line segments has the same end point.
    # If the right-hand matrix is not invertible, then two line segments are parallel.
    # If none of above conditions is matched, find the solution of s and t,
    # if both s and t are in [0, 1], then two line segments intersect.

    v0 = line_a[1] - line_a[0]
    v1 = line_b[1] - line_b[0]
    det = v0.x * v1.y - v0.y * v1.x
    # Two line segments are parallel
    if det == 0:
        # TODO Determine if two lines overlap
        return False

    du = line_a[0] - line_b[0]
    s_det = v1.x * du.y - v1.y * du.x
    t_det = v0.x * du.y - v0.y * du.x

    if ((det > 0 and 0 <= s_det <= det and 0 <= t_det <= det) or
        (det < 0 and det <= s_det <= 0 and det <= t_det <= 0)):
        return True

    return False

def rect_collideline(rect: Rect, line) -> bool:
    """
    Check if line segment intersects with a rect

    @param rect The Rect of the target rectangle
    @param line A tuple (Vector2, Vector2) representing both end points
           of line segment
    """
    # Either of line ends is in the target rect.
    rect_expanded = rect.inflate(1, 1)  # Take the bottom and right line into account
    if rect_expanded.collidepoint(line[0]) or rect_expanded.collidepoint(line[1]):
        return True

    line_top = (Vector2(rect.topleft), Vector2(rect.topright))
    line_bottom = (Vector2(rect.bottomleft), Vector2(rect.bottomright))
    line_left = (Vector2(rect.topleft), Vector2(rect.bottomleft))
    line_right = (Vector2(rect.topright), Vector2(rect.bottomright))

    return (line_intersect(line_top, line) or
        line_intersect(line_bottom, line) or
        line_intersect(line_left, line) or
        line_intersect(line_right, line))

def rect_break_or_contact_box(rect: Rect, box: Rect):
    """
    Determine if the `rect` breaks the `box` or it contacts the border of `box`

    @param rect The Rect of the target rectangle
    @param box The target box
    """
    return (
        rect.left <= box.left or
        rect.right >= box.right or
        rect.top <= box.top or
        rect.bottom >= box.bottom)

def bounce_off_ip(bounce_obj_rect: Rect, bounce_obj_speed,
    hit_obj_rect: Rect, hit_obj_speed):
    """
    Calculate the speed and position of the `bounce_obj` after it bounces off the `hit_obj`.
    The position of `bounce_obj_rect` and the value of `bounce_obj_speed` will be updated.

    This function should be called only when two objects are colliding.

    @param bounce_obj_rect The Rect of the bouncing object
    @param bounce_obj_speed The 2D speed vector of the bouncing object.
    @param hit_obj_rect The Rect of the hit object
    @param hit_obj_speed The 2D speed vector of the hit object
    """
    # Treat the hit object as an unmovable object
    speed_diff_x = bounce_obj_speed[0] - hit_obj_speed[0]
    speed_diff_y = bounce_obj_speed[1] - hit_obj_speed[1]

    # The relative position between top and bottom, and left and right
    # of two objects at the last frame
    rect_diff_bT_hB = hit_obj_rect.bottom - bounce_obj_rect.top + speed_diff_y
    rect_diff_bB_hT = hit_obj_rect.top - bounce_obj_rect.bottom + speed_diff_y
    rect_diff_bL_hR = hit_obj_rect.right - bounce_obj_rect.left + speed_diff_x
    rect_diff_bR_hL = hit_obj_rect.left - bounce_obj_rect.right + speed_diff_x

    # Get the surface distance from the bouncing object to the hit object
    # and the new position for the bouncing object if it really hit the object
    # according to their relative position
    ## The bouncing object is at the bottom
    if rect_diff_bT_hB < 0 and rect_diff_bB_hT < 0:
        surface_diff_y = rect_diff_bT_hB
        extract_pos_y = hit_obj_rect.bottom
    ## The bouncing object is at the top
    elif rect_diff_bT_hB > 0 and rect_diff_bB_hT > 0:
        surface_diff_y = rect_diff_bB_hT
        extract_pos_y = hit_obj_rect.top - bounce_obj_rect.height
    else:
        surface_diff_y = -1 if speed_diff_y > 0 else 1

    ## The bouncing object is at the right
    if rect_diff_bL_hR < 0 and rect_diff_bR_hL < 0:
        surface_diff_x = rect_diff_bL_hR
        extract_pos_x = hit_obj_rect.right
    ## The bouncing object is at the left
    elif rect_diff_bL_hR > 0 and rect_diff_bR_hL > 0:
        surface_diff_x = rect_diff_bR_hL
        extract_pos_x = hit_obj_rect.left - bounce_obj_rect.width
    else:
        surface_diff_x = -1 if speed_diff_x > 0 else 1

    # Calculate the duration to hit the surface for x and y coordination.
    # 計算先碰到物體的左右 還是先碰到物體上下
    time_hit_y = surface_diff_y / speed_diff_y
    time_hit_x = surface_diff_x / speed_diff_x

    if time_hit_y >= 0 and time_hit_y >= time_hit_x:
        # 先碰到y座標，移動物體的y就會變成對應的y
        bounce_obj_speed[1] *= -1
        bounce_obj_rect.y = extract_pos_y

    if time_hit_x >= 0 and time_hit_y <= time_hit_x:
        # 先碰到x座標，移動物體的x就會變成對應的x
        bounce_obj_speed[0] *= -1
        bounce_obj_rect.x = extract_pos_x

def bounce_off(bounce_obj_rect: Rect, bounce_obj_speed,
    hit_obj_rect: Rect, hit_obj_speed):
    """
    The alternative version of `bounce_off_ip`. The function returns the result
    instead of updating the value of `bounce_obj_rect` and `bounce_obj_speed`.

    @return A tuple (`new_bounce_obj_rect`, `new_bounce_obj_speed`)
    """
    new_bounce_obj_rect = bounce_obj_rect.copy()
    new_bounce_obj_speed = bounce_obj_speed.copy()

    bounce_off_ip(new_bounce_obj_rect, new_bounce_obj_speed,
        hit_obj_rect, hit_obj_speed)

    return new_bounce_obj_rect, new_bounce_obj_speed

def bounce_in_box_ip(bounce_obj_rect: Rect, bounce_obj_speed,
    box_rect: Rect):
    """
    Bounce the object if it hits the border of the box.
    The speed and the position of the `bounce_obj` will be updated.

    @param bounce_obj_rect The Rect of the bouncing object
    @param bounce_obj_speed The 2D speed vector of the bouncing object.
    """
    if bounce_obj_rect.left <= box_rect.left:
        bounce_obj_rect.left = box_rect.left
        bounce_obj_speed[0] *= -1
    elif bounce_obj_rect.right >= box_rect.right:
        bounce_obj_rect.right = box_rect.right
        bounce_obj_speed[0] *= -1

    if bounce_obj_rect.top <= box_rect.top:
        bounce_obj_rect.top = box_rect.top
        bounce_obj_speed[1] *= -1
    elif bounce_obj_rect.bottom >= box_rect.bottom:
        bounce_obj_rect.bottom = box_rect.bottom
        bounce_obj_speed[1] *= -1

def bounce_in_box(bounce_obj_rect: Rect, bounce_obj_speed,
    box_rect: Rect):
    """
    The alternative version of `bounce_in_box_ip`. The function returns the result
    instead of updating the value of `bounce_obj_rect` and `bounce_obj_speed`.

    @return A tuple (new_bounce_obj_rect, new_bounce_obj_speed)
    """
    new_bounce_obj_rect = bounce_obj_rect.copy()
    new_bounce_obj_speed = bounce_obj_speed.copy()

    bounce_in_box_ip(new_bounce_obj_rect, new_bounce_obj_speed, box_rect)

    return (new_bounce_obj_rect, new_bounce_obj_speed)



import pygame
import time

def quit_or_esc() -> bool:
    """
    Check if the quit event is triggered or the ESC key is pressed.
    """
    for event in pygame.event.get():
        if (event.type == pygame.QUIT or
            (event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE)):
            return True
    return False

class KeyCommandMap:
    """
    Map the keys to the commands and return the mapped command when the corresponding
    key is pressed
    """
    def __init__(self, command_map: dict):
        """
        Constructor

        @param command_map A dict which maps the keys to the commands.
               The key of the dict is the key-code defined in pygame, and
               the value is the command that will be returned when the corresponding
               key is pressed.
        """
        if not isinstance(command_map, dict):
            raise TypeError("The 'command_map' should be a 'dict'.")

        self._command_map = command_map

    def get_pressed_commands(self):
        """
        Check the pressed keys and return the corresponding commands

        @return A list of commands of which corresponding keys are pressed
                If there is no registered key pressed, return an empty list.
        """
        key_pressed_list = pygame.key.get_pressed()
        pressed_commands = []
        for key, command in self._command_map.items():
            if key_pressed_list[key]:
                pressed_commands.append(command)

        return pressed_commands

class FPSCounter:
    """
    The counter for calculating the FPS

    Invoke `get_FPS()` at each frame. The counter will count how many calls within
    a specified updating interval. Within a updating interval, the returned FPS value
    won't be updated until the starting of next updating interval.
    """

    def __init__(self, update_interval = 1.0):
        """
        Constructor

        @param update_interval The time interval in seconds for updating the FPS value
        """
        self._update_interval = update_interval
        self._fps = 0
        self._tick_count = 0
        self._last_time_updated = time.time()

    def get_FPS(self) -> int:
        """
        Update and get the calculated FPS
        """
        self._tick_count += 1

        current_time = time.time()
        if current_time - self._last_time_updated > self._update_interval:
            self._fps = int(round(self._tick_count / (current_time - self._last_time_updated)))
            self._tick_count = 0
            self._last_time_updated = current_time

        return self._fps


from os import path
import random
import pygame.draw
from mlgame.utils.enum import get_ai_name
from mlgame.view.view_model import create_asset_init_data, create_image_view_data, create_rect_view_data

from .env import TURN_LEFT_CMD, TURN_RIGHT_CMD, FORWARD_CMD, BACKWARD_CMD, \
    AIM_LEFT_CMD, AIM_RIGHT_CMD, SHOOT, SHOOT_COOLDOWN, IMAGE_DIR, ORANGE, BLUE, IS_DEBUG
from .Gun import Gun
from .GenerateMap import MapGenerator

Vec = pygame.math.Vector2


class Player(pygame.sprite.Sprite):
    def __init__(self, construction, **kwargs):
        super().__init__()
        """
        初始化玩家資料
        construction可直接由TiledMap打包地圖資訊後傳入
        :param construction:
        :param kwargs:
        """
        self.id = construction["_id"]
        self.no = construction["_no"]
        self.rect = pygame.Rect(construction["_init_pos"], construction["_init_size"])
        self.play_rect_area = kwargs["play_rect_area"]
        self.origin_xy = self.rect.topleft
        self.origin_center = self.rect.center
        self.origin_size = (self.rect.width, self.rect.height)
        self.original_rect = self.rect.copy()
        self.draw_pos = self.rect.topleft
        self.surface = pygame.Surface(self.origin_size)
        self.angle = 0
        self.score = 0
        self.used_frame = 0
        self.last_shoot_frame = 0
        self.lives = 3
        self.power = 10
        self.vel = Vec(0, 0)
        self.speed = 8 * construction["_init_size"][0] / 50
        # TODO refactor use vel
        self.move_dict = {"left_up": Vec(-self.speed, -self.speed), "right_up": Vec(self.speed, -self.speed),
                     "left_down": Vec(-self.speed, self.speed), "right_down": Vec(self.speed, self.speed),
                     "left": Vec(-self.speed, 0), "right": Vec(self.speed, 0), "up": Vec(0, -self.speed),
                     "down": Vec(0, self.speed)}
        self.rot = 0
        self.last_shoot_frame = self.used_frame
        self.last_turn_frame = self.used_frame
        self.rot_speed = 45
        self.oil = 100
        self.is_alive = True
        self.is_shoot = False
        self.is_forward = False
        self.is_backward = False
        self.is_turn_right = False
        self.is_turn_left = False
        self.collided = False
        self.action_history = []
        self.act_cd = kwargs["act_cd"]
        self.quadrant = 0

        self.gun = Gun(self.id, self.rect.topleft, (self.rect.width, self.rect.height), **kwargs)

        self.calculate_quadrant()

    def calculate_quadrant(self):
        mid_x = self.play_rect_area.width // 2
        mid_y = (self.play_rect_area.height - 100) // 2
        self.quadrant = (
            1 if self.rect.x >= mid_x and self.rect.y < mid_y else
            2 if self.rect.x < mid_x and self.rect.y < mid_y else
            3 if self.rect.x < mid_x and self.rect.y >= mid_y else
            4
        )

    def update(self, command: dict):
        self.used_frame += 1
        if self.lives <= 0:
            self.is_alive = False
            self.gun.is_alive = False
            self.lives = 0

        if not self.is_alive:
            self.rect.topleft = Vec(self.play_rect_area.left + self.origin_size[0]*self.no
                                    , self.play_rect_area.top - self.origin_size[1]*2)
            return

        self.rotate()
        self.gun.update(self.rect.center)

        if not self.act_cd:
            self.is_turn_right = False
            self.is_turn_left = False
        elif self.used_frame - self.last_turn_frame > self.act_cd:
            self.is_turn_right = False
            self.is_turn_left = False

        if self.rect.right > self.play_rect_area.right \
                or self.rect.left < self.play_rect_area.left \
                or self.rect.bottom > self.play_rect_area.bottom \
                or self.rect.top < self.play_rect_area.top:
            self.collide_with_walls()
        else:
            self.act(command[get_ai_name(self.no - 1)])

    def rotate(self):
        self.rot = self.rot % 360
        self.angle = 3.14 / 180 * self.rot
        new_sur = pygame.transform.rotate(self.surface, self.rot)
        origin_center = self.rect.center
        self.rect = new_sur.get_rect()
        self.rect.center = origin_center
        self.draw_pos = self.rect.topleft

    def act(self, commands: list):
        if not commands or self.collided:
            return None

        # Only one action from the three categories can be executed in one frame
        move_flag = False
        shoot_flag = False
        aim_flag = False

        while commands:
            command = commands.pop()

            # Shoot
            if not shoot_flag:
                if self.power and SHOOT == command:
                    self.shoot()

                shoot_flag = True

            # Aiming
            # TODO: Maybe the oil should be consumed when aiming
            if not aim_flag:
                if command == AIM_LEFT_CMD:
                    self.gun.turn_left()
                elif command == AIM_RIGHT_CMD:
                    self.gun.turn_right()

                aim_flag = True

            if self.oil <= 0:
                self.oil = 0
                continue

            # Movement
            if not move_flag:
                if TURN_LEFT_CMD == command and not self.is_turn_left and TURN_RIGHT_CMD != command:
                    self.oil -= 0.1
                    self.turn_left()
                    self.is_turn_left = True
                    self.is_forward = False
                    self.is_backward = False
                    self.is_turn_right = False
                elif TURN_RIGHT_CMD == command and not self.is_turn_right and TURN_LEFT_CMD != command:
                    self.oil -= 0.1
                    self.turn_right()
                    self.is_turn_right = True
                    self.is_forward = False
                    self.is_backward = False
                    self.is_turn_left = False
                elif FORWARD_CMD == command and BACKWARD_CMD != command:
                    self.oil -= 0.1
                    self.forward()
                    self.is_forward = True
                    self.is_backward = False
                    self.is_turn_right = False
                    self.is_turn_left = False
                    self.action_history.append(FORWARD_CMD)
                elif BACKWARD_CMD == command and FORWARD_CMD != command:
                    self.oil -= 0.1
                    self.backward()
                    self.is_backward = True
                    self.is_forward = False
                    self.is_turn_right = False
                    self.is_turn_left = False
                    self.action_history.append(BACKWARD_CMD)

                move_flag = True

            self.action_history = self.action_history[-1:]

    def shoot(self):
        if self.last_shoot_frame == 0 or self.used_frame - self.last_shoot_frame > SHOOT_COOLDOWN:
            self.last_shoot_frame = self.used_frame
            self.power -= 1
            self.is_shoot = True

    def forward(self):
        if self.id != 1:
            rot = self.rot + 180
            if rot >= 360:
                rot -= 360
        else:
            rot = self.rot
        if rot == 0:
            self.rect.center += self.move_dict["left"]
        elif rot == 315:
            self.rect.center += self.move_dict["left_up"]
        elif rot == 270:
            self.rect.center += self.move_dict["up"]
        elif rot == 225:
            self.rect.center += self.move_dict["right_up"]
        elif rot == 180:
            self.rect.center += self.move_dict["right"]
        elif rot == 135:
            self.rect.center += self.move_dict["right_down"]
        elif rot == 90:
            self.rect.center += self.move_dict["down"]
        elif rot == 45:
            self.rect.center += self.move_dict["left_down"]

    def backward(self):
        if self.id != 1:
            rot = self.rot + 180
            if rot >= 360:
                rot -= 360
        else:
            rot = self.rot
        if rot == 0:
            self.rect.center += self.move_dict["right"]
        elif rot == 315:
            self.rect.center += self.move_dict["right_down"]
        elif rot == 270:
            self.rect.center += self.move_dict["down"]
        elif rot == 225:
            self.rect.center += self.move_dict["left_down"]
        elif rot == 180:
            self.rect.center += self.move_dict["left"]
        elif rot == 135:
            self.rect.center += self.move_dict["left_up"]
        elif rot == 90:
            self.rect.center += self.move_dict["up"]
        elif rot == 45:
            self.rect.center += self.move_dict["right_up"]

    def turn_left(self):
        self.last_turn_frame = self.used_frame
        self.rot += self.rot_speed

    def turn_right(self):
        self.last_turn_frame = self.used_frame
        self.rot -= self.rot_speed

    def collide_with_walls(self):
        # Retrieve last two actions
        last_actions = self.action_history[-1:]

        # Reverse actions
        for action in reversed(last_actions):
            if action == FORWARD_CMD:
                self.backward()
            elif action == BACKWARD_CMD:
                self.forward()

    def collide_with_bullets(self):
        self.lives -= 1

    def get_power(self, power: int):
        self.power += power
        if self.power > 10:
            self.power = 10
        elif self.power < 0:
            self.power = 0

    def get_oil(self, oil: int):
        self.oil += oil
        if self.oil > 100:
            self.oil = 100
        elif self.oil < 0:
            self.oil = 0

    def get_rot(self):
        if self.id == 2:
            return (self.rot + 180) % 360
        return self.rot

    def get_data_from_obj_to_game(self) -> dict:
        info = {"id": f"{self.no}P"
                , "x": self.rect.x
                , "y": self.rect.y
                , "speed": self.speed
                , "score": self.score
                , "power": self.power
                , "oil": self.oil
                , "lives": self.lives
                , "angle": self.get_rot()
                , "gun_angle": self.gun.get_rot()
                , "cooldown": 0
                    if self.last_shoot_frame == 0
                    or self.used_frame - self.last_shoot_frame > SHOOT_COOLDOWN
                    else SHOOT_COOLDOWN - self.used_frame + self.last_shoot_frame,
                }
        return info

    def get_obj_progress_data(self) -> dict:
        if not self.is_alive:
            return []
        image_data = create_image_view_data(f"{self.id}P", *self.draw_pos, *self.origin_size, self.angle)
        return image_data

    def get_obj_init_data(self) -> list:
        img_data = {"1P": "https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/1P_body.svg",
                    "2P": "https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/2P_body.svg"}
        image_init_data = []
        for id, url in img_data.items():
            image_init_data.append(create_asset_init_data(id, self.origin_size[0], self.origin_size[1],
                                                          path.join(IMAGE_DIR, f"{id}_body.png"), url))
        return image_init_data

    def get_info_to_game_result(self) -> dict:
        info = {"no": f"{self.no}P"
                , "score": self.score
                , "lives": self.lives
                }
        if IS_DEBUG:
            if self.rect.right > self.play_rect_area.right \
                    or self.rect.left < self.play_rect_area.left \
                    or self.rect.bottom > self.play_rect_area.bottom \
                    or self.rect.top < self.play_rect_area.top:
                print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!【OUT】!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            else:
                print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!【SAFE】!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        return info



import pygame.sprite


def get_size(sprite: pygame.sprite.Sprite):
    return sprite.rect.width, sprite.rect.height


def set_topleft(sprite: pygame.sprite.Sprite, top_left: tuple):
    sprite.rect.topleft = top_left


def add_score(sprite: pygame.sprite.Sprite, score: int):
    sprite.score += score


def set_shoot(sprite: pygame.sprite.Sprite, is_shoot: bool):
    sprite.is_shoot = is_shoot


def get_sprites_progress_data(sprites: pygame.sprite.Group):
    data_list = []
    for sprite in sprites:
        data = sprite.get_obj_progress_data()
        if data:
            data_list.append(data)
    return data_list



from os import path

import pygame
from mlgame.view.view_model import (create_asset_init_data,
                                    create_image_view_data)

from .env import IMAGE_DIR


class Gun(pygame.sprite.Sprite):
    def __init__(self, id, pos, size, **kwargs):
        super().__init__()
        self.id = id
        self.rect = pygame.Rect(pos, size)
        self.origin_size = (self.rect.width, self.rect.height)
        self.draw_pos = self.rect.topleft
        self.surface = pygame.Surface(self.origin_size)
        self.rot = 0
        self.rot_speed = 45

        self.used_frame = 0
        self.last_turn_frame = self.used_frame
        self.act_cd = kwargs["act_cd"]

        self.is_alive = True
        self.is_turn_left = False
        self.is_turn_right = False

        if self.id == 1:
            self.pivot_offset = pygame.Vector2(-8, 0)
        else:
            self.pivot_offset = pygame.Vector2(8, 0)

    def update(self, gun_pos):
        self.used_frame += 1
        self.rotate()

        if not self.act_cd:
            self.is_turn_right = False
            self.is_turn_left = False
        elif self.used_frame - self.last_turn_frame > self.act_cd:
            self.is_turn_right = False
            self.is_turn_left = False

        self.rect.center = gun_pos + self.pivot_offset.rotate(-self.rot)
        self.draw_pos = self.rect.topleft

    def rotate(self):
        self.rot = self.rot % 360
        self.angle = 3.14 / 180 * self.rot
        new_sur = pygame.transform.rotate(self.surface, self.rot)
        origin_center = self.rect.center
        self.rect = new_sur.get_rect()
        self.rect.center = origin_center
        self.draw_pos = self.rect.topleft

    def turn_left(self):
        if self.is_turn_left:
            return

        self.rot += self.rot_speed
        self.last_turn_frame = self.used_frame

        self.is_turn_left = True
        self.is_turn_right = False

    def turn_right(self):
        if self.is_turn_right:
            return

        self.rot -= self.rot_speed
        self.last_turn_frame = self.used_frame

        self.is_turn_left = False
        self.is_turn_right = True

    def get_rot(self):
        if self.id == 2:
            return (self.rot + 180) % 360
        return self.rot

    def get_obj_progress_data(self) -> dict:
        if not self.is_alive:
            return {}
        image_data = create_image_view_data(
            f"{self.id}P_gun", *self.draw_pos, *self.origin_size, self.angle
        )
        return image_data

    def get_obj_init_data(self) -> list:
        img_data = {
            "1P_gun": "https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/1P_gun.svg",
            "2P_gun": "https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/2P_gun.svg",
        }
        image_init_data = []
        for id, url in img_data.items():
            image_init_data.append(
                create_asset_init_data(
                    id,
                    self.origin_size[0],
                    self.origin_size[1],
                    path.join(IMAGE_DIR, f"{id}.png"),
                    url,
                )
            )
        return image_init_data



import math

from os import path
import pygame

NAME = "name"
TYPE = "type"
ANGLE = "angle"
SIZE = "size"
COLOR = "color"
CORDINATE = "coordinate"
IMAGE = "image"
RECTANGLE = "rect"
POLYGON = "polygon"

'''data path'''
IMAGE_DIR = path.join(path.dirname(__file__), 'image')


def trnsfer_hex_to_rgb(hex):
    h = hex.lstrip('#')
    return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))


class PygameView():
    def __init__(self, game_info: dict):
        pygame.display.init()
        self.scene_init_data = game_info
        self.width = self.scene_init_data["scene"]["width"]
        self.height = self.scene_init_data["scene"]["height"]
        self.background_color = trnsfer_hex_to_rgb(self.scene_init_data["scene"][COLOR])
        self.screen = pygame.display.set_mode((self.width, self.height))
        self.address = "GameView"
        self.image_dict = self.loading_image()
        self.font = {}
        self.map_width = game_info["map_width"]
        self.map_height = game_info["map_height"]
        self.pygame_point = [500 - self.map_width, 480 - self.map_height]
        # if "images" in game_info.keys():
        #     self.image_dict = self.loading_image(game_info["images"])

    def loading_image(self):
        result = {}
        for file in self.scene_init_data["assets"]:
            if file[TYPE] == IMAGE:
                image = pygame.image.load(path.join(IMAGE_DIR, file["image_id"] + ".png"))
                result[file["image_id"]] = image
        return result

    def draw(self, object_imformation):
        '''
        每個frame呼叫一次，把角色畫在螢幕上
        :param all_sprite:
        :return:
        '''
        self.draw_screen()
        self.limit_pygame_screen()
        for game_object in object_imformation["game_object_list"]:
            if game_object[TYPE] == IMAGE:
                self.draw_image(game_object["image_id"], game_object["x"], game_object["y"],
                                game_object["width"], game_object["height"], game_object["angle"])

            elif game_object[TYPE] == RECTANGLE:
                self.draw_rect(game_object["x"], game_object["y"], game_object["width"], game_object["height"],
                               trnsfer_hex_to_rgb(game_object[COLOR]))

            elif game_object[TYPE] == POLYGON:
                self.draw_polygon(game_object["points"], trnsfer_hex_to_rgb(game_object[COLOR]))

            elif game_object[TYPE] == "text":
                self.draw_text(game_object["content"], game_object["font-style"],
                               game_object["x"], game_object["y"], trnsfer_hex_to_rgb(game_object[COLOR]))
            elif game_object[TYPE] == "line":
                self.draw_line(game_object["x1"], game_object["y1"], game_object["x2"], game_object["y2"],
                               game_object["width"], game_object[COLOR])

            else:
                pass

        for game_object in object_imformation["game_background"]:
            if game_object[TYPE] == IMAGE:
                self.draw_image(game_object["image_id"], game_object["x"] - self.pygame_point[0],
                                game_object["y"] - self.pygame_point[1],
                                game_object["width"], game_object["height"], game_object["angle"])

            elif game_object[TYPE] == RECTANGLE:
                self.draw_rect(game_object["x"] - self.pygame_point[0], game_object["y"] - self.pygame_point[1],
                               game_object["width"], game_object["height"],
                               trnsfer_hex_to_rgb(game_object[COLOR]))

            elif game_object[TYPE] == "text":
                self.draw_text(game_object["content"], game_object["font-style"],
                               game_object["x"] - self.pygame_point[0], game_object["y"] - self.pygame_point[1],
                               trnsfer_hex_to_rgb(game_object[COLOR]))
            else:
                pass

    def draw_screen(self):
        self.screen.fill(self.background_color)  # hex # need turn to RGB

    def draw_image(self, image_id, x, y, width, height, angle):
        image = pygame.transform.rotate(pygame.transform.scale(self.image_dict[image_id], (width, height)),
                                        (angle * 180 / math.pi) % 360)
        rect = image.get_rect()
        rect.x, rect.y = x + self.pygame_point[0], y + self.pygame_point[1]
        self.screen.blit(image, rect)

    def draw_rect(self, x, y, width, height, color):
        pygame.draw.rect(self.screen, color,
                         pygame.Rect(x + self.pygame_point[0], y + self.pygame_point[1], width, height))

    def draw_line(self, x1, y1, x2, y2, width, color):
        pygame.draw.line(self.screen, color, (x1 + self.pygame_point[0], y1 + self.pygame_point[1]),
                         (x2 + self.pygame_point[0], y2 + self.pygame_point[1]), width)

    def draw_polygon(self, points, color):
        vertices = []
        for p in points:
            vertices.append((p["x"] + self.pygame_point[0], p["y"] + self.pygame_point[1]))
        pygame.draw.polygon(self.screen, color, vertices)

    def flip(self):
        pygame.display.flip()

    def draw_text(self, text, font_style, x, y, color):
        if font_style in self.font.keys():
            font = self.font[font_style]
        else:
            list = font_style.split(" ", -1)
            size = int(list[0].replace("px", "", 1))
            font_type = list[1].lower()
            font = pygame.font.Font(pygame.font.match_font(font_type), size)
            self.font[font_style] = font
        text_surface = font.render(text, True, color)
        text_rect = text_surface.get_rect()
        text_rect.x, text_rect.y = (x + self.pygame_point[0], y + self.pygame_point[1])
        self.screen.blit(text_surface, text_rect)

    def limit_pygame_screen(self):

        keystate = pygame.key.get_pressed()
        if keystate[pygame.K_w]:
            self.pygame_point[1] += 2
        elif keystate[pygame.K_s]:
            self.pygame_point[1] -= 2
        elif keystate[pygame.K_a]:
            self.pygame_point[0] += 2
        elif keystate[pygame.K_d]:
            self.pygame_point[0] -= 2

        if self.pygame_point[1] < 480 - self.map_height:
            self.pygame_point[1] = 480 - self.map_height
        elif self.pygame_point[1] > 0:
            self.pygame_point[1] = 0
        else:
            pass
        if self.pygame_point[0] < 500 - self.map_width:
            self.pygame_point[0] = 500 - self.map_width
        elif self.pygame_point[0] > 0:
            self.pygame_point[0] = 0
        else:
            pass



import datetime
import json
from enum import Enum, auto

import pydantic as pydantic


class ProcessError(Exception):
    """
    The base class for the exception occurred in the process
    """

    def __init__(self, process_name, message):
        """
        Constructor

        @param process_name The name of the process in which the error occurred
        @param message The error message
        """
        self.process_name = process_name
        self.message = message

class GameProcessError(ProcessError):
    """
    Exception raised when an error occurred in the game process
    """
    pass

class MLProcessError(ProcessError):
    """
    Exception raised when an error occurred in the ml process
    """
    pass

def trim_callstack(exception_msg: str, target_user_file: str):
    """
    Shorten the call stack to the starting point of the user script
    """
    exception_msg_list = exception_msg.splitlines(keepends = True)
    # Store title
    trimmed_msg = exception_msg_list[0]

    # Find the starting point
    i = 0
    for i in range(2, len(exception_msg_list)):
        if target_user_file in exception_msg_list[i]:
            break

    return trimmed_msg + "".join(exception_msg_list[i:])

class ExecutionCommandError(Exception):
    """
    Exception raised when parsed invalid execution command
    """
    def __init__(self, message):
        """
        Constructor
        """
        self.message = message

    def __str__(self):
        return self.message

class GameConfigError(Exception):
    """
    Exception raised when the game provides invalid game config
    """
    def __init__(self, message):
        """
        Constructor
        """
        self.message = message

    def __str__(self):
        return self.message


class ErrorEnum(str, Enum):
    AI_INIT_ERROR = auto()
    AI_EXEC_ERROR = auto()
    GAME_EXEC_ERROR = auto()
    COMMAND_ERROR = auto()


class GameError(pydantic.BaseModel):
    error_type: ErrorEnum
    message: str = ""
    frame: int
    time_stamp: datetime.datetime = datetime.datetime.now(datetime.timezone.utc)

    def data(self):
        data = {
            "type": "game_error",
            "data": {
                "error_type": self.error_type.name,
                "frame": self.frame,
                "message": self.message,
                "time_stamp": self.time_stamp.isoformat()
            }
        }
        return json.dumps(data)



from mlgame.utils.logger import logger
# 安全鎖：禁止所有 exec*（subprocess.run 最終會被擋）
def install_exec_killer():
    import sys
    logger.info(f"您的作業系統是: {sys.platform}")
    if sys.platform == "win32":
        return  # Windows 不適用 seccomp
    elif sys.platform == "darwin":
        return  # MacOS 不適用 seccomp
    
    logger.info("您的作業系統是 Linux，安裝 exec killer 以防止 execve 被使用")

    import pyseccomp as sc
    from pyseccomp import Arch
    import errno
    # 可選：先設 no_new_privs，避免繞過（多數容器已預設開）
    try:
        import ctypes
        libc = ctypes.CDLL(None)
        PR_SET_NO_NEW_PRIVS = 38
        libc.prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0)
    except Exception:
        pass

    try:
        if hasattr(sc, "SCMP_ACT_ALLOW"):  # A 類：libseccomp 綁定（python3-libseccomp）
            logger.info("libseccomp 綁定，禁止使用 execve")
            flt = sc.SyscallFilter(defaction=sc.SCMP_ACT_ALLOW)
            flt.add_rule(sc.SCMP_ACT_ERRNO, sc.ScmpSyscall("execve"))
            # flt.add_rule(sc.SCMP_ACT_ERRNO, sc.ScmpSyscall("execveat"))
            # （如需要可再加 posix_spawn/posix_spawnp，視 libc 而定）
            flt.load()
        else:  # B 類：pyseccomp 綁定（pip: pyseccomp）
            logger.info("pyseccomp 綁定，禁止使用 execve")
            flt = sc.SyscallFilter(defaction=sc.ALLOW)
            
            # flt.add_rule(sc.ERRNO(errno.EPERM), "execveat")
            flt.add_rule(sc.ERRNO(errno.EPERM), "execve")
            flt.load()
    except Exception as e:
        logger.exception(f"Exception: {e}")


# guard.py
import sys

def lock_exec_linux_with_seccomp():
    # 僅 Linux 才嘗試 seccomp；macOS 直接跳過
    if sys.platform != "linux":
        return False
    try:
        import seccomp as sc
    except Exception:
        return False  # 沒裝 pyseccomp，就回報沒上鎖

    # 安裝 seccomp：禁止 execve/execveat
    if hasattr(sc, "SCMP_ACT_ALLOW"):
        flt = sc.SyscallFilter(defaction=sc.SCMP_ACT_ALLOW)
        flt.add_rule(sc.SCMP_ACT_ERRNO, sc.ScmpSyscall("execve"))
        flt.add_rule(sc.SCMP_ACT_ERRNO, sc.ScmpSyscall("execveat"))
        flt.load()
    else:
        flt = sc.SyscallFilter(defaction=sc.ALLOW)
        flt.add_rule(sc.ERRNO, "execve")
        flt.add_rule(sc.ERRNO, "execveat")
        flt.load()
    return True

def lock_exec_userspace_fallback():
    # macOS / 無 seccomp 時的替代方案（可被繞過，開發/測試用）
    import builtins, subprocess, os

    # 1) PEP 578 audit hook：攔常見事件
    try:
        def _audit_hook(event, args):
            if event in ("subprocess.Popen", "os.exec", "os.posix_spawn"):
                raise PermissionError("subprocess/exec is blocked by policy")
            # 可選：也擋 socket 連線
            if event in ("socket.__init__", "socket.connect"):
                raise PermissionError("network is blocked by policy")
        sys.addaudithook(_audit_hook)
    except Exception:
        pass

    # 2) 保險：monkey patch 常用 API（避免沒觸發 audit 的路徑）
    def _blocked(*a, **k):
        raise PermissionError("blocked by policy")
    for name in ("Popen", "call", "check_call", "check_output", "run", "getoutput", "getstatusoutput"):
        if hasattr(subprocess, name):
            setattr(subprocess, name, _blocked)
    for name in ("execl", "execle", "execlp", "execlpe", "execv", "execve", "execvp", "execvpe", "posix_spawn", "posix_spawnp"):
        if hasattr(os, name):
            setattr(os, name, _blocked)

def harden_startup():
    # 1) 盡量用 fork，避免 spawn/exec
    if sys.platform != "win32":
        import multiprocessing as mp
        try:
            mp.set_start_method("fork", force=True)
        except RuntimeError:
            pass

    # 2) 先嘗試 Linux seccomp；失敗就用 fallback
    if not lock_exec_linux_with_seccomp():
        lock_exec_userspace_fallback()


"""
to deprecated
"""
import pickle
import time

from pathlib import Path
from mlgame.argument.execution_command import GameMode

RECORD_FORMAT_VERSION = 2

def get_recorder(execution_cmd, ml_names):
    """
    The helper function for generating a recorder object
    """
    if not execution_cmd.record_progress:
        return DummyRecorder()

    root_dir_path = Path(__file__).parent.parent
    log_dir_path = root_dir_path.joinpath(
        "games", execution_cmd.game_name, "log")

    game_params_str = [str(p) for p in execution_cmd.game_params]
    filename_prefix = (
        "manual" if execution_cmd.game_mode == GameMode.MANUAL else "ml")
    if game_params_str:
        filename_prefix += "_" + "_".join(game_params_str)

    return Recorder(ml_names, log_dir_path, filename_prefix)

class Recorder:
    """
    Record the scene information and the game command to the file
    """
    def __init__(
            self, ml_names: list, saving_directory: Path,
            filename_prefix: str = ""):
        """
        Constructor

        @param ml_names A list containing the name of all ml clients
        @param saving_directory Specify the directory for saving files
        @param filename_prefix Specify the prefix of the filename to be generated.
               The filename will be "<prefix>_YYYY-MM-DD_hh-mm-ss.pickle".
        """
        self._saving_directory = saving_directory
        if not self._saving_directory.exists():
            self._saving_directory.mkdir()

        if not isinstance(filename_prefix, str):
            raise TypeError("'filename_prefix' should be the type of 'str'")
        self._filename_prefix = filename_prefix

        # Create storing slots for each ml client
        game_progress = {
            "record_format_version": RECORD_FORMAT_VERSION
        }
        for name in ml_names:
            game_progress[name] = {
                "scene_info": [],
                "command": []
            }
        self._game_progress = game_progress
        self._ml_names = ml_names

    def record(self, scene_info_dict: dict, cmd_dict: dict):
        """
        Record the scene information and the command

        The received scene information will be stored in a list.

        @param scene_info_dict A dict storing the scene information for each client
        @param cmd_dict A dict storing the command received from each client
        """
        for name in self._ml_names:
            target_slot = self._game_progress[name]
            scene_info = scene_info_dict.get(name, None)
            if scene_info:
                target_slot["scene_info"].append(scene_info)
                target_slot["command"].append(cmd_dict.get(name, None))

    def flush_to_file(self):
        """
        Flush the stored objects to the file
        """
        filename = time.strftime("%Y-%m-%d_%H-%M-%S") + ".pickle"

        if self._filename_prefix:
            filename = self._filename_prefix + "_" + filename

        filepath = self._saving_directory.joinpath(filename)
        with open(filepath, "wb") as f:
            pickle.dump(self._game_progress, f)

        for name in self._ml_names:
            target_slot = self._game_progress[name]
            target_slot["scene_info"].clear()
            target_slot["command"].clear()

class DummyRecorder:
    """
    The recorder that only proivdes the API of `Recorder` but do nothing
    """
    def __init__(self):
        pass

    def record(self, scene_info, commands):
        pass

    def flush_to_file(self):
        pass



import functools
import time


# Decorator to measure the execution time of a function in milliseconds
def timeit(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        execution_time_ms = (end - start) * 1000  # Convert to milliseconds
        print(f"{func.__name__} executed in {execution_time_ms:.4f} ms")
        return result

    return wrapper


# Decorator to measure the execution time of a function in milliseconds using perf_counter
def timeit_in_perf(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = time.perf_counter()
        result = func(*args, **kwargs)
        end = time.perf_counter()
        execution_time_ms = (end - start) * 1000  # Convert to milliseconds
        print(f"{func.__name__} executed in {execution_time_ms:.4f} ms")
        return result

    return wrapper



import pygame
import json
from .env import PPM, HEIGHT


# Map 讀取地圖資料
class TiledMap_box2d:
    def __init__(self, filename: str, TILESIZE):
        pygame.init()
        self.data = []
        with open(filename) as json_file:
            data = json.load(json_file)
            self.tileHeight = data["height"]
            self.tileWidth = data["width"]
            map_data = data["layers"][0]["data"]
            map = []
            # for i in range(self.tileHeight):
            #     self.data.append(data["layers"][0]["data"][i * self.tileWidth:(i + 1) * self.tileWidth])
            for i in range(len(map_data)):
                if map_data[i]:
                    self.data.append([i // self.tileWidth, i % self.tileWidth, map_data[i]])
        self.width = self.tileWidth * TILESIZE
        self.height = self.tileHeight * TILESIZE
        self.PPM = 20
        self.screen = pygame

    def get_wall_info(self):  # 擷取迷宮
        wall_tiles = []
        passed = []
        j = 0
        first_tile = -1
        last_tile = -1  # 前一顆tile
        type = ''
        while len(passed) < len(self.data):
            if j in passed:
                j += 1
                if j == len(self.data):
                    j = 0
                continue
            if self.data[j][2] == 1:
                if first_tile == -1:  # a new wall
                    first_tile = j  # index
                    last_tile = j
                    passed.append(j)
                    j += 1
                    if (self.data[j][0] == self.data[last_tile][0] and
                            self.data[j][1] == self.data[last_tile][1] + 1 and self.data[j][2] == 1):  # 檢查橫向有沒有延伸的積木
                        type = 'h'
                    else:
                        type = 'v'
                    continue
                elif type == 'h':  # 橫向積木
                    if (self.data[j][0] == self.data[last_tile][0] and
                            self.data[j][1] == self.data[last_tile][1] + 1):
                        last_tile = j
                        passed.append(j)
                        if j == len(self.data) - 1:
                            wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                            j = 0
                            first_tile = -1
                            continue
                        else:
                            j += 1
                            continue
                    else:
                        wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                        j = 0
                        first_tile = -1
                elif type == 'v':  # 縱向積木
                    if (self.data[j][1] == self.data[last_tile][1] and
                            self.data[j][0] == self.data[last_tile][0] + 1):
                        last_tile = j
                        passed.append(j)
                        if j == len(self.data) - 1:
                            wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                            j = 0
                            first_tile = -1
                            continue
                        else:
                            j += 1
                            continue
                    elif (self.data[j][0] > self.data[last_tile][0] + 1 and self.data[j][1] > self.data[last_tile][1]):
                        wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                        j = 0
                        first_tile = -1
                    else:
                        j += 1
                        if j == len(self.data) - 1:
                            wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                            j = 0
                            first_tile = -1
                            continue
                        continue
                else:
                    continue
            else:
                passed.append(j)
                j += 1
                continue
        return wall_tiles

    def transfer_to_box2d(self, wall: list):
        '''
        :param wall:[[0 ,0 , 1],[0, 29, 1]]
        :return: vertices of box2d body
        (y, height - 1 - x) = box2d中左下角的點的座標
        '''
        vertices = []
        if wall[0][0] == wall[1][0]:  # 橫向積木
            left_tile_x = wall[0][1]
            left_tile_y = self.tileHeight - 1 - wall[0][0]
            right_tile_x = wall[1][1]
            right_tile_y = self.tileHeight - 1 - wall[1][0]
            vertices.extend([[left_tile_x, left_tile_y], [left_tile_x, left_tile_y + 1],
                             [right_tile_x + 1, right_tile_y], [right_tile_x + 1, right_tile_y + 1]])
        else:  # 縱向積木
            upper_tile_x = wall[0][1]
            upper_tile_y = self.tileHeight - 1 - wall[0][0]
            down_tile_x = wall[1][1]
            down_tile_y = self.tileHeight - 1 - wall[1][0]
            vertices.extend([[upper_tile_x, upper_tile_y + 1], [upper_tile_x + 1, upper_tile_y + 1],
                             [down_tile_x, down_tile_y], [down_tile_x + 1, down_tile_y]])
        return vertices

    def load_other_obj(self):
        obj = {"car": None,
               "end_point": None,
               "check_point": []}
        for i in range(len(self.data)):
            if self.data[i][2] == 1:
                continue
            elif self.data[i][2] in range(2, 6):
                # obj["car"] = self.data[i]
                # from tiled -> box2D
                obj["car"] = [self.tileHeight - 1 - self.data[i][0], self.data[i][1], self.data[i][2]]
            elif self.data[i][2] == 6:
                obj["end_point"] = [self.tileHeight - self.data[i][0], self.data[i][1], self.data[i][2]]
                # obj["end_point"] = self.data[i]
            elif self.data[i][2] == 7:
                obj["check_point"].append([self.tileHeight - 1 - self.data[i][0], self.data[i][1], self.data[i][2]])
                # obj["check_point"].append(self.data[i])
        return obj

    def get_check_wall_info(self):
        end = []
        for i in self.data:
            if i[2] == 7:
                end.append(self.data.index(i))

        wall_tiles = []
        passed = []
        j = 0
        first_tile = -1
        last_tile = -1  # 前一顆tile
        type = ''
        while len(passed) < len(self.data):
            if j in passed:
                j += 1
                if j == len(self.data):
                    j = 0
                continue
            if self.data[j][2] == 7:
                if first_tile == -1:  # a new wall
                    first_tile = j  # index
                    last_tile = j
                    passed.append(j)
                    end.remove(j)
                    j += 1
                    if (self.data[j][0] == self.data[last_tile][0] and
                            self.data[j][1] == self.data[last_tile][1] + 1 and
                            self.data[j][2] == 7):  # 檢查橫向有沒有延伸的積木
                        type = 'h'
                    else:
                        type = 'v'
                    continue
                elif type == 'h':  # 橫向積木
                    if (self.data[j][0] == self.data[last_tile][0] and
                            self.data[j][1] == self.data[last_tile][1] + 1):
                        last_tile = j
                        passed.append(j)
                        if j == end[-1]:
                            wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                            end.remove(j)
                            j = 0
                            first_tile = -1
                            continue
                        else:
                            end.remove(j)
                            j += 1
                            continue
                    else:
                        wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                        j = 0
                        first_tile = -1
                elif type == 'v':  # 縱向積木
                    if (self.data[j][1] == self.data[last_tile][1] and
                            self.data[j][0] == self.data[last_tile][0] + 1):
                        last_tile = j
                        passed.append(j)
                        if j == end[-1]:
                            wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                            end.remove(j)
                            j = 0
                            first_tile = -1
                            continue
                        else:
                            end.remove(j)
                            j += 1
                            continue
                    elif (self.data[j][0] > self.data[last_tile][0] + 1):
                        wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                        j = 0
                        first_tile = -1
                    elif j == end[-1]:
                        wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                        j = 0
                        first_tile = -1
                        continue
                    else:
                        j += 1
                        if j == end[-1]:
                            wall_tiles.append([self.data[first_tile], self.data[last_tile]])
                            end.remove(j)
                            j = 0
                            first_tile = -1
                            continue
                        continue
                else:
                    continue
            else:
                passed.append(j)
                if j == len(self.data) - 1:
                    j = 0
                    continue
                j += 1
                continue
        return wall_tiles

    def print_data(self):
        print("width", self.width)
        print("height", self.height)
        print("tile_width", self.tileWidth)
        print("tile_height", self.tileHeight)
        for i in range(len(self.data)):
            print(self.data[i])



from enum import Enum, auto
from functools import lru_cache


@lru_cache(16)
def get_ai_name(user_index: int = 0):
    return f"{user_index + 1}P"


class StringEnum(str, Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name

    def __eq__(self, other):
        if isinstance(other, StringEnum):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == other

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.value

    def __hash__(self):
        return hash(self.value)



from loguru import logger

from mlgame.argument.cmd_argument import parse_cmd_and_get_arg_obj
from mlgame.argument.model import GroupAI, GroupEnum
from .env import GAMES_PATH, AI_CLIENTS_PATH, OUTPUT_PATH, get_path


def test_parse_output_folder_arg():
    arg_str = f"-o {OUTPUT_PATH} -i {get_path(AI_CLIENTS_PATH, 'arkanoid/rule/ml_play.py')} {get_path(GAMES_PATH, 'arkanoid')} "
    args = parse_cmd_and_get_arg_obj(arg_str.split(' '))

    assert "var" in str(args.output_folder)
    assert isinstance(args.group_ai[0], GroupAI)
    assert not args.is_manual


def test_parse_input_ai_arg():
    arg_str = f"-i {get_path(AI_CLIENTS_PATH, 'arkanoid/rule/ml_play.py')} {get_path(GAMES_PATH, 'arkanoid')} "
    args = parse_cmd_and_get_arg_obj(arg_str.split(' '))

    assert isinstance(args.group_ai[0], GroupAI)
    assert args.group_ai[0].ai_label == "1P"

    ai_label = "ai_label"
    arg_str = f"-i {get_path(AI_CLIENTS_PATH, 'arkanoid/rule/ml_play.py')},{ai_label} {get_path(GAMES_PATH, 'arkanoid')} "
    args = parse_cmd_and_get_arg_obj(arg_str.split(' '))

    assert isinstance(args.group_ai[0], GroupAI)
    assert args.group_ai[0].ai_label == ai_label


# def test_parse_cmd_with_group_ai_and_input_ai():
#     arg_str = (f"--group_ai A,1P,{get_path(AI_CLIENTS_PATH, 'arkanoid/dev/ml_play.py')},kylin2 " +
#                f"--group_ai B,2P,{get_path(AI_CLIENTS_PATH, 'arkanoid/dev/ml_play.py')},kylin3 " +
#                f"--group_ai B,3P,{get_path(AI_CLIENTS_PATH, 'arkanoid/dev/ml_play.py')},kylin4 " +
#                f"{get_path(GAMES_PATH, 'arkanoid')} ")
#     args = parse_cmd_and_get_arg_obj(arg_str.split(' '))
#
#     assert len(args.group_ai) == 3
#     assert isinstance(args.group_ai[0], GroupAI)
#     assert not args.is_manual
#     logger.info(args.group_ai)

#
# def test_parse_group_ai_cmd():
#     arg_str = (f"--group_ai A,1P,{get_path(AI_CLIENTS_PATH, 'arkanoid/dev/ml_play.py')},ai_label_1 " +
#                f"--group_ai A,2P,{get_path(AI_CLIENTS_PATH, 'arkanoid/dev/ml_play.py')},ai_label_2 " +
#                f"--group_ai B,3P,{get_path(AI_CLIENTS_PATH, 'arkanoid/dev/ml_play.py')},ai_label_3 " +
#                f"--group_ai B,4P,{get_path(AI_CLIENTS_PATH, 'arkanoid/dev/ml_play.py')},ai_label_4 " +
#                f"{get_path(GAMES_PATH, 'arkanoid')} ")
#
#     args = parse_cmd_and_get_arg_obj(arg_str.split(' '))
#
#     assert len(args.group_ai) == 4
#     assert isinstance(args.group_ai[0], GroupAI)
#     assert not args.is_manual
#
#     logger.warning(args.group_ai)


def test_enum():
    group = "A"
    assert group == GroupEnum.A
    assert group == GroupEnum(group)
    assert group == GroupEnum.A.value
    assert group != str(GroupEnum.A)





import pygame
from os import path
from .env import *

class SoundController():
    def __init__(self, is_sound_on):
        if is_sound_on == "on":
            self.is_sound_on = True
            try:
                pygame.mixer.init()
                self.hit_sound = pygame.mixer.Sound(path.join(SOUND_DIR, "explosion.wav"))
                self.coin_sound = pygame.mixer.Sound(path.join(SOUND_DIR, "coin.wav"))
                self.lose_sound = pygame.mixer.Sound(path.join(SOUND_DIR,"Powerup3.wav"))
                pygame.mixer.music.load(path.join(SOUND_DIR, "BGM.mp3"))
                pygame.mixer.music.set_volume(0.4)
            except Exception:
                self.is_sound_on = False

        else:
            self.is_sound_on = False

    def play_music(self):
        if self.is_sound_on:
            pygame.mixer.music.play(-1)
        else:
            pass

    def play_hit_sound(self):
        if self.is_sound_on:
            self.hit_sound.play()
        else:
            pass

    def play_coin_sound(self):
        if self.is_sound_on:
            self.coin_sound.play()
        else:
            pass

    def play_lose_sound(self):
        if self.is_sound_on:
            self.lose_sound.play()
        else:
            pass



import random
from .math_function import *
import pygame
import Box2D
from .env import *


class Sensor():
    def __init__(self, world, body, sensor_num, angle):
        self.car = body
        self.right_value = {}
        self.right_top_value = {}
        self.left_value = {}
        self.left__top_value = {}
        self.front_value = {}
        self.sensor_num = sensor_num
        angle_t = int(angle / 0.5) % 4
        self.sensor_right = world.CreateDynamicBody(
            position=(body.position[0] + sensor_trans[angle_t][0], body.position[1] + sensor_trans[angle_t][1]))
        ball = self.sensor_right.CreateCircleFixture(radius=0.1)
        self.sensor_left = world.CreateDynamicBody(
            position=(body.position[0] - sensor_trans[angle_t][0], body.position[1] - sensor_trans[angle_t][1]))
        ball = self.sensor_left.CreateCircleFixture(radius=0.1)

        world.CreateDistanceJoint(bodyA=self.sensor_left, bodyB=body, anchorA=self.sensor_left.position,
                                  anchorB=body.position, collideConnected=True)
        world.CreateDistanceJoint(bodyA=self.sensor_right, bodyB=body, anchorA=self.sensor_right.position,
                                  anchorB=body.position, collideConnected=True)
        self.last_detect_sensor = -3
        pass

    def update(self, frame, walls):
        self.front_value = self.sensor_detect(walls, (0, 1))
        self.right_value = self.sensor_detect(walls, (1, 0))
        self.left_value = self.sensor_detect(walls, (-1, 0))
        if self.sensor_num == 5:
            self.right_top_value = self.sensor_detect(walls, (1, 1))
            self.left_top_value = self.sensor_detect(walls, (-1, 1))
        else:
            self.right_top_value = {"coordinate": self.car.position, "distance": -1}
            self.left_top_value = {"coordinate": self.car.position, "distance": -1}

        self.last_detect_sensor = frame
        if frame - self.last_detect_sensor > 2:
            self.sensor_right.position = self.car.GetWorldVector(localVector=(1, 0)) + self.car.position
            self.sensor_left.position = self.car.GetWorldVector(localVector=(-1, 0)) + self.car.position
            pass
        return {"right_value": self.right_value,
                "left_value": self.left_value,
                "front_value": self.front_value,
                "right_top_value": self.right_top_value,
                "left_top_value": self.left_top_value}

    def sensor_detect(self, walls, vector):
        car_center = self.car.position
        distance = []
        results = []
        dots = []
        vector = self.car.GetWorldVector(localVector=vector)
        sensor_value = {}

        for wall in walls:
            distance.append(cross_point_dot(car_center,
                                            vector,
                                            wall[0], wall[1]))
        for dot in distance:
            if dot:
                if dot[0] - car_center[0] > 0 and vector[0] > 0:
                    results.append(math.sqrt(
                        (dot[0] - car_center[0]) ** 2 + (dot[1] - car_center[1]) ** 2) - 1)
                    dots.append(dot)
                elif dot[0] - car_center[0] < 0 and vector[0] < 0:
                    results.append(math.sqrt(
                        (dot[0] - car_center[0]) ** 2 + (dot[1] - car_center[1]) ** 2) - 1)
                    dots.append(dot)
                else:
                    pass
            else:
                pass

        try:
            coordinate = dots[results.index(min(results))]
            sensor_value = {"coordinate": (round(coordinate[0], 3), round(coordinate[1], 3)),
                            "distance": round(min(results) * 5 * random.uniform(0.95, 1.05), 1)
                            }
            # self.front_value = round(min(results) * 5 * random.uniform(0.95, 1.05), 1)
            if sensor_value["distance"] < 0:
                sensor_value["distance"] = 0
            return sensor_value
        except TypeError:
            sensor_value["distance"] = -1
            sensor_value["coordinate"] = self.car.position
            return sensor_value
        except ValueError:
            sensor_value["distance"] = -1
            sensor_value["coordinate"] = self.car.position
            return sensor_value
